part of '../export.dart';

/// Base interfaces for classes that know how to
/// - set a value
/// - get that value synchronously
/// - subscribe to a wave to receive new values.
abstract class PureSea<V> {
  Wave<V> get wave;

  void add(V v);

  V get value;
}

abstract class Sea<V> extends Wave<V>
    with WaveType<V>, WaveOtherMixin<V>, WaveOperators<V>
    implements PureSea<V> {

  factory Sea(V v) => _Sea(v);

  Sea.wave(WaveType<V> wave) : super(wave);

  void change(V Function(V) action) {
    add(action(value));
  }

  SeaLens<V, R> then<R>(Lens<V, R> lens) {
    return SeaLens.rawLens(this, lens);
  }

  Sea<R> transform<R>(Lens<V, R> lens) {
    return _AnonymousSea(
      this.map(lens.get),
          () => lens.get(value),
          (val) => this.add(lens.update(this.value, val)),
    );
  }
}

class _AnonymousSea<V> extends Sea<V> {
  @protected
  final void Function(V) _add;

  @override
  void add(V v) => _add(v);

  final V Function() __value;

  @override
  V get value => __value();

  @override
  final Wave<V> wave;

  _AnonymousSea(this.wave, this.__value, this._add) : super.wave(wave);
}

/// A [Wave] with a synchronously available and changeable value.
/// Meant to reduce Signal boilerplate.
///
/// It doesn't implement Sink to suppress the close-sinks
/// lint warning that Signals have.
///
/// A [Sea] is also a [Wave] so that it is backwards compatible with consumers
/// expecting a [Wave].
///
/// Add is only meant to be called inside of HookBlocs. That's why it's set
/// as protected.
class _Sea<V> extends Sea<V> {
  final Future<void> Function() _close;

  @override
  final Wave<V> wave;

  /// .add is supposed to be called in HookBloc subclasses never outside.
  /// You're supposed to add this line to the top of your HookBloc.
  /// You might also want to ignore type declarations because they get inferred.
  // ignore_for_file: invalid_use_of_protected_member
  // ignore_for_file: type_annotate_public_apis
  @protected
  final void Function(V) _add;

  @override
  void add(V v) => _add(v);

  final V Function() __value;

  @override
  V get value => __value();

  factory _Sea(V s) {
    final signal = Signal(s);
    return _Sea._create(
          () async => signal.close(),
      signal.wave,
      signal.add,
          () => signal.value,
    );
  }

  _Sea._create(this._close, this.wave, this._add, this.__value)
      : super.wave(wave);

  /// Use this in combination with [SeaHookBlocExtension.d] to dispose this
  /// [Sea] and hide the implementation details.
  Sea<V> operator &(
      _Sea<V> Function(_Sea<V> self, void Function() dispose) action) {
    return action(this, _close);
  }
}

/// Experimental High level HookBloc extensions for boilerplate free Blocs.
extension SeaHookBlocExtension on HookBloc {
  Sea<T> d<T>(_Sea<T> t, void Function() dispose) {
    onDispose.add(() async => dispose());
    return t;
  }
}

extension SeaExtension<T> on T {
  /// Use this inside of [HookBloc]s as a replacement for [HookBloc.disposeSink].
  Sea<T> get $ {
    final sea = _Sea(this);
    HookBloc.disposeEffect(sea, (_Sea<T> a) async => a._close());
    return sea;
  }

  /// This is meant to be use when initialized inside of constructors.
  /// Use the & operator and [d] extension method on [HookBloc]s to dispose
  /// this [Sea].
  ///
  /// ```
  /// final a = 1.sea & d;
  /// ```
  ///
  _Sea<T> get $$ {
    return _Sea(this);
  }
}

extension Tuple2ABC<A, B> on Tuple2<A, B> {
  A get a => value1;

  B get b => value2;
}

extension Tuple3ABC<A, B, C> on Tuple3<A, B, C> {
  A get a => value1;

  B get b => value2;

  C get c => value3;
}

extension Tuple4ABC<A, B, C, D> on Tuple4<A, B, C, D> {
  A get a => value1;

  B get b => value2;

  C get c => value3;

  D get d => value4;
}

extension Tuple5ABC<A, B, C, D, E> on Tuple5<A, B, C, D, E> {
  A get a => value1;

  B get b => value2;

  C get c => value3;

  D get d => value4;

  E get e => value5;
}

extension Tuple6ABC<A, B, C, D, E, F> on Tuple6<A, B, C, D, E, F> {
  A get a => value1;

  B get b => value2;

  C get c => value3;

  D get d => value4;

  E get e => value5;

  F get f => value6;
}

extension Tuple7ABC<A, B, C, D, E, F, G> on Tuple7<A, B, C, D, E, F, G> {
  A get a => value1;

  B get b => value2;

  C get c => value3;

  D get d => value4;

  E get e => value5;

  F get f => value6;

  G get g => value7;
}

extension TExtensions<T> on T {
  R apply<R>(R Function(T) action) => action(this);
}