import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:meta/meta.dart';

abstract class JSONPrototypeCodec<T extends SmartX<T, dynamic>>
    implements PrototypeCodec<T, dynamic> {
  @override
  Codec<T, dynamic> codec(T prototype);

  Codec<dynamic, String> toStringCodec(T prototype);
}

/// Deserialization of maps can have missing values for certain keys.
/// This describes the behaviour of such cases.
///
/// This is not to be confused with values that are explicitly set to null.
/// Nulls are valid in JSON and therefore maps. This handles only the cases
/// where the value was expected to be contained in the source map but
/// is missing completely (map.contains(key) == false).
enum MapMissingBehaviour {

  /// The value is kept the way it is.
  /// It will probably be the default value but in cases
  /// where the to be converted object is mutated before
  /// the conversion one could maybe want to keep the
  /// previously mutated values.
  ignore,

  /// The value is set to null.
  setNull,
}

/// Converts multiple values each with a key to a json map.
///
/// Needs: list of [JSONSmartVariableInterface] variables with String keys.
///
/// Uses their keys as the map keys
/// and uses their serialization as the values..
///
/// If a map is missing a value [missingBehaviour] handles how these cases
/// should be handled.
///
/// By default, nulls are added to the resulting map in [encodeDirectly].
/// Therefore if [missingBehaviour] is [MapMissingBehaviour.ignore]
/// [encodeDirectly] and [decodeWithPrototype] will not guarantee to be inverses of each other.
class JSONMapTactic<T extends SmartJSON<T>> implements JSONPrototypeCodec<T> {

  final Map<String, JSONSmartVariableInterface> structure;

  /// If you choose to nest tactics then you'll need to set this.
  ///
  /// Because a tactic takes the root tactic of [T] by default it can't
  /// be nested without telling it which tactic to take.
  ///
  /// In case of nested tactics let this tactic know where it is in [T].
  final JSONMapTactic<T> Function(T) tacticChooser;

  final MapMissingBehaviour missingBehaviour;

  @visibleForTesting
  const JSONMapTactic.raw(this.structure,
      this.tacticChooser,
      this.missingBehaviour)
      : assert(structure != null),
        assert(tacticChooser != null),
        assert(missingBehaviour != null)
  ;

  /// Some values might just not have a default value that would make sense
  /// Or if data from a foreign api is received there could also be a benefit in removing
  /// missing values by setting them to null.
  factory JSONMapTactic.nullOnMissing(
      Map<String, JSONSmartVariableInterface> structure,
      {JSONMapTactic<T> Function(T) tacticChooser}) =>
      JSONMapTactic.raw(
        structure,
        tacticChooser ?? ((t) => t.tactic() as JSONMapTactic<T>),
        MapMissingBehaviour.setNull,
      );

  /// If a value is not present during decoding then a members current value is used.
  factory JSONMapTactic.ignoreMissing(
      Map<String, JSONSmartVariableInterface> structure,
      {JSONMapTactic<T> Function(T) tacticChooser}) =>
      JSONMapTactic.raw(
        structure,
        tacticChooser ?? ((t) => t.tactic() as JSONMapTactic<T>),
        MapMissingBehaviour.ignore,
      );

  Map<String, dynamic> rawPrimitiveMap() =>
      structure.map<String, dynamic>((key, value) =>
          MapEntry<String, dynamic>(
            key,
            value.codec().decode(value.nullVal()),
          ));

  @override
  Codec<T, dynamic> codec(T prototype) =>
      AnonymousCodec<T, dynamic>(
        AnonymousConverter<T, dynamic>((T t) {
          return tacticChooser(t).rawPrimitiveMap();
        }),
        AnonymousConverter<dynamic, T>((dynamic data) {
          if (data != null) {
            final m = Map<dynamic, dynamic>.from(data as Map).cast<
                String,
                dynamic>();
            tacticChooser(prototype).structure.forEach((key, value) {
              if (m.containsKey(key)) {
                value.safeSet(value.codec().encode(m[key]));
              } else {
                switch (missingBehaviour) {
                  case MapMissingBehaviour.ignore:
                    break;
                  case MapMissingBehaviour.setNull:
                    value.safeSet(value.codec().encode(null));
                    break;
                }
              }
            });
          }
          return prototype;
        }),
      );

  @override
  Codec<Object, String> toStringCodec(T prototype) {
    return json;
  }
}

class JSONCanonicalMapTactic<T extends SmartJSON<T>>
    implements JSONPrototypeCodec<T> {

  final Map<String, JSONSmartVariableInterface> structure;

  final JSONCanonicalMapTactic<T> Function(T) tacticChooser;

  final Map<String, JSONSmartVariableInterface> toJsonStructure;

  @visibleForTesting
  const JSONCanonicalMapTactic.raw(this.structure,
      this.tacticChooser,
      this.toJsonStructure)
      : assert(structure != null),
        assert(tacticChooser != null)
  ;

  /// Lets you provide a different canonical map for the [decodeWithPrototype] stage.
  ///
  /// This constructor is equivalent to the [JSONMapTactic.ignoreMissing] constructor
  /// but provides a solution for the following problem:
  ///
  /// Example:
  /// One source may provide A: {"a": 123}
  /// A different source may provide B: {"A": 123}
  ///
  /// If it's impossible to tell whether A or B is used but
  /// their values are always identical then use this constructor
  /// with a fromJson definition like this
  ///
  ///     {"a": x, "A": x}
  factory JSONCanonicalMapTactic.ignoreOnMissingCanonical(
      Map<String, JSONSmartVariableInterface> structure,
      Map<String, JSONSmartVariableInterface> toJsonMap,
      {JSONCanonicalMapTactic<T> Function(T) tacticChooser}) =>
      JSONCanonicalMapTactic<T>.raw(
        structure,
        tacticChooser ?? ((t) => t.tactic() as JSONCanonicalMapTactic<T>),
        toJsonMap,
      );

  @override
  Codec<T, dynamic> codec(T prototype) =>
      AnonymousCodec<T, dynamic>(
        AnonymousConverter<T, dynamic>((T t) {
          final a = t.tactic() as JSONCanonicalMapTactic<T>;
          return (a.toJsonStructure ?? a.structure).map<String, dynamic>((key,
              value) =>
              MapEntry<String, dynamic>(
                key,
                value.codec().decode(value.nullVal()),
              ));
        }),
        AnonymousConverter<dynamic, T>((dynamic data) {
          final m = Map<dynamic, dynamic>.from(data as Map).cast<String,
              dynamic>();
          tacticChooser(prototype).structure.forEach((key, value) {
            if (m.containsKey(key)) {
              value.safeSet(value.codec().encode(m[key]));
            }
          });
          return prototype;
        }),
      );

  @override
  Codec<Object, String> toStringCodec(T prototype) {
    return json;
  }
}

/// Converts one value to its single json representation.
class JSONSingleTactic<T extends SmartJSON<T>>
    implements JSONPrototypeCodec<T> {

  final JSONSmartVariableInterface smartVariable;

  const JSONSingleTactic(this.smartVariable);

  @override
  Codec<T, Object> codec(T prototype) {
    return AnonymousCodec(
      AnonymousConverter((T t) =>
          (t.tactic() as JSONSingleTactic<T>).smartVariable.to()),
      AnonymousConverter((Object data) {
        (prototype.tactic() as JSONSingleTactic<T>).smartVariable.safeSet(
            smartVariable.codec().encode(data));
        return prototype;
      }),
    );
  }

  @override
  Codec<Object, String> toStringCodec(T prototype) {
    return json;
  }
}

/// Converts a list of values to a json list with their index inside
/// [variables] being the index in the converted list and vice versa.
///
/// The keys dictate the index inside the list.
class JSONListTactic<T extends SmartJSON<T>> extends JSONPrototypeCodec<T> {

  final List<JSONSmartVariableInterface> variables;

  /// If you choose to nest tactics then you'll need to set this.
  ///
  /// Because a tactic takes the root tactic of [T] by default it can't
  /// be nested without telling it which tactic to take.
  ///
  /// In case of nested tactics let this tactic know where it is in [T].
  final JSONListTactic<T> Function(T) tacticChooser;

  JSONListTactic(this.variables, {JSONListTactic<T> Function(T) tacticChooser})
      : tacticChooser = tacticChooser ?? ((t) =>
  t.tactic() as JSONListTactic<T>);

  @override
  Codec<T, Object> codec(T prototype) {
    return AnonymousCodec(
      AnonymousConverter((T t) {
        return (t.tactic() as JSONListTactic<T>).variables.map<dynamic>((
            value) {
          return value.codec().decode(value.nullVal());
        }).toList();
      }),
      AnonymousConverter((Object data) {
        final m = List<dynamic>.from(data as Iterable);
        tacticChooser(prototype).variables.asMap().forEach((key, value) {
          if (key < m.length) {
            value.safeSet(value.codec().encode(m[key]));
          } else {
            /// Same problem as with [JSONMapTactic]
            /// Do [MapMissingBehaviour.ignore] for now
            /// and add others if needed.
          }
        });
        return prototype;
      }),
    );
  }

  @override
  Codec<Object, String> toStringCodec(T prototype) {
    return json;
  }
}

/// For maps that are treated as lists where
/// - the key represents a value
/// - the keys all have the same type
/// - the values all have the same type
/// Example:
///
/// {"0.123": "0.134",  "0.314": "431.34"}
///
/// Puts all keys and values each into a [JSONSmartVariableInterface].
class JSONMapAsListTactic<T extends SmartJSON<T>, K, V>
    implements JSONPrototypeCodec<T> {

  final JSONSmartVariableInterface<Iterable<K>> keys;

  final JSONSmartVariableInterface<Iterable<V>> values;

  const JSONMapAsListTactic(this.keys, this.values);

  Map<K, V> asMap() {
    final k = keys.option | null;
    final v = values.option | null;
    if (k != null && v != null) {
      assert(k.length == v.length);
      final _keysList = k.toList();
      final _valueList = v.toList();
      return _keysList.asMap().map((index, key) =>
          MapEntry(key, _valueList[index]));
    } else {
      return {};
    }
  }

  @override
  Codec<T, Object> codec(T prototype) {
    return AnonymousCodec(
      AnonymousConverter((T t) {
        return Map<dynamic, dynamic>.fromIterables(
            keys.to() as List, values.to() as List);
      }),
      AnonymousConverter((Object data) {
        final from = data as Map;
        final tactic = prototype.tactic() as JSONMapAsListTactic<T, K, V>;
        tactic.keys.safeSet(tactic.keys.codec().encode(from?.keys?.toList()));
        tactic.values.safeSet(
            tactic.values.codec().encode(from?.values?.toList()));
        return prototype;
      }),
    );
  }

  @override
  Codec<Object, String> toStringCodec(T prototype) => json;
}