import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart' as z;
import 'package:decimal/decimal.dart';
import 'package:quiver/collection.dart';
import 'package:trie/trie.dart';

/// A list for primitive types:
///
/// - int
/// - double
/// - bool
/// - String
///
/// Or set a listConverter to convert from json to Lists of T.
/// Example:
///
/// var t = SmartListT<_SmartEnumTestEnum>.converter(
///    [_SmartEnumTestEnum.one],
///    listConverter: (value) {
///        return List.of(value).map((f) => firstOfList(_SmartEnumTestEnum.values, _SmartEnumTestEnum.one)(f?.toString())).toList();
///    },
///    useToStringToJson: true,
/// );
class SmartListT<T> extends JSONSmartVariableInterface<List<T>> {
  final List<T> Function(dynamic) listConverter;
  final bool useToStringToJson;

  SmartListT(List<T> init, {this.useToStringToJson = false})
      : listConverter = ((dynamic a) =>
  a != null ?
  List<dynamic>.from(a as List)
      .toList()
      .cast<T>() : null),
        super(init);

  SmartListT.converter(List<T> init,
      {this.listConverter, this.useToStringToJson = false})
      : super(init);

  SmartListT.enums(List<T> init, T deflt, List<T> allT)
      :
        useToStringToJson = true,
        listConverter = ((dynamic value) =>
            List<dynamic>.of(value as List).map((dynamic f) =>
                firstOfList(allT, deflt)(f?.toString())).toList()),
        super(init);

  void applyToggle(T value) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    if (a.contains(value)) {
      a.remove(value);
    } else {
      a.add(value);
    }

    safeSet(a);
  }

  void applyAdd(T value) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    a.add(value);
    safeSet(a);
  }

  void applyRemove(T value) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    a.remove(value);
    return safeSet(a);
  }

  void applyRemoveAtIndex(int index) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    if (a.length > index) {
      a.removeAt(index);
      return safeSet(a);
    }
  }

  void applyReorderIndexTo(int fromIndex, int toIndex) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();

    if (a.length > fromIndex && a.length > toIndex) {
      final pivot = a.removeAt(fromIndex);
      a.insert(toIndex, pivot);
      safeSet(a);
    }
  }

  void applyReplaceAtIndex(int index, T t) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    if (a.length > index) {
      a[index] = t;
      return safeSet(a);
    }
  }

  void applyApplyAtIndex(int index, T Function(T) t) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    if (a.length > index) {
      a[index] = t(a[index]);
      return safeSet(a);
    }
  }

  void applyReplaceWhere(T Function(T) t, bool Function(T t) where) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    safeSet(a.map((current) {
      if (where(current)) {
        return t(current);
      } else {
        return current;
      }
    }).toList());
  }

  T getAtIndexOr(int index, {T Function() orElse}) {
    return option.fold(() => orElse?.call(), (a) {
      if (a.length > index) {
        return a[index];
      } else {
        return orElse?.call();
      }
    });
  }

  void append(z.Option<T> t) {
    safeSet(t.map((val) => [...nullVal(), val]) | []);
  }

  void applyRemoveWhere(bool Function(T) where) {
    final a = List<dynamic>.from(nullVal() ?? <dynamic>[]).cast<T>();
    a.removeWhere(where);
    return safeSet(a);
  }

  @override
  Codec<dynamic, List<T>> codec() =>
      IsJSONSmartCodec(
        fromJSON: (dynamic a) => listConverter(a),
        useToStringToJson: useToStringToJson,
      );
}

class SmartModelList<T extends SmartJSON<T>> extends SmartListT<T> {
  SmartModelList(List<T> init, T Function() creator)
      : super.converter(
    init ?? [],
    listConverter: (dynamic a) =>
        List<dynamic>.from((a ?? <T>[]) as List).map(creator().from)
            .toList()
            .cast<T>(),
  );
}

class SmartListTWithMultiMap<K, T> extends SmartListT<T> {

  final K Function(T) keyMap;

  Multimap<K, T> _map;

  SmartListTWithMultiMap(List<T> init, this.keyMap,
      {bool useToStringToJson = false})
      : super(init, useToStringToJson: useToStringToJson);

  SmartListTWithMultiMap.converter(List<T> init, this.keyMap,
      {List<T> Function(dynamic) listConverter, bool useToStringToJson = false})
      : super.converter(
      init, listConverter: listConverter, useToStringToJson: useToStringToJson);

  @override
  void set(z.Option<List<T>> t) {
    super.set(t);
    _map = Multimap.fromIterable(t | [], key: (dynamic a) => keyMap(a as T),
        value: (dynamic a) => a as T);
  }

  Iterable<T> mapGet(K key) => _map[key];

  bool mapContains(K key) => _map.containsKey(key);

  @override
  Codec<dynamic, List<T>> codec() =>
      IsJSONSmartCodec(
        fromJSON: (dynamic a) => listConverter(a),
        useToStringToJson: useToStringToJson,
      );
}

class SmartModelListWithMultiMap<K, T extends SmartJSON<T>>
    extends SmartListTWithMultiMap<K, T> {
  SmartModelListWithMultiMap(List<T> init, K Function(T) keyMap,
      T Function() creator)
      : super.converter(
    init ?? [],
    keyMap,
    listConverter: (dynamic a) =>
        List<dynamic>.from((a ?? <T>[]) as List).map(creator().from)
            .toList()
            .cast<T>(),
  );
}

class SmartListTWithStartsWithEndsWithTrie<T>
    extends SmartListTWithMultiMap<String, T> {

  // https://stackoverflow.com/questions/21521729/how-do-i-reverse-a-string-in-dart
  static String _reverse(String s) {
    final length = s.length;
    final charCodes = List<int>(length);

    for (var index = 0; index < length; index++) {
      charCodes[index] = s.codeUnitAt(length - index - 1);
    }
    return String.fromCharCodes(charCodes);
  }

  Trie _startsWithTrie;
  Trie _endsWithTrie;

  SmartListTWithStartsWithEndsWithTrie(List<T> init, String Function(T) keyMap,
      {bool useToStringToJson = false})
      : super(init, keyMap, useToStringToJson: useToStringToJson);

  SmartListTWithStartsWithEndsWithTrie.converter(List<T> init,
      String Function(T) keyMap,
      {List<T> Function(dynamic) listConverter, bool useToStringToJson = false})
      : super.converter(init, keyMap, listConverter: listConverter,
      useToStringToJson: useToStringToJson);

  @override
  void set(z.Option<List<T>> t) {
    super.set(t);
    final list = t.map((List<T> a) => a.map(keyMap).toList()) | [];
    _startsWithTrie = Trie.list(list);
    _endsWithTrie = Trie.list(list.map(_reverse).toList());
    _map = Multimap.fromIterable(t | [], key: (dynamic a) => keyMap(a as T),
        value: (dynamic a) => a as T);
  }

  Iterable<T> getStartsWith(String str) {
    assert(str != null);
    return _startsWithTrie.getAllWordsWithPrefix(str)
        .map((a) => _map[a])
        .expand((a) => a);
  }

  Iterable<T> getEndsWith(String str) {
    assert(str != null);
    return _endsWithTrie.getAllWordsWithPrefix(_reverse(str)).map((
        a) => _map[_reverse(a)]).expand((a) => a);
  }

  @override
  Codec<dynamic, List<T>> codec() =>
      IsJSONSmartCodec(
        fromJSON: (dynamic a) => listConverter(a),
        useToStringToJson: useToStringToJson,
      );
}

class SmartModelListWithWithStartsWithEndsWithTrie<T extends SmartJSON<T>>
    extends SmartListTWithStartsWithEndsWithTrie<T> {
  SmartModelListWithWithStartsWithEndsWithTrie(List<T> init,
      String Function(T) keyMap, T Function() creator)
      : super.converter(
    init ?? [],
    keyMap,
    listConverter: (dynamic a) =>
        List<dynamic>.from((a ?? <T>[]) as List).map(creator().from)
            .toList()
            .cast<T>(),
  );
}

class SmartDecimalList extends SmartListT<Decimal> {
  SmartDecimalList(List<Decimal> init) : super.converter(init);

  @override
  Codec<dynamic, List<Decimal>> codec() => _ListDecimalSmartConverter();
}

class _ListDecimalSmartConverter extends Codec<dynamic, List<Decimal>> {
  @override
  Converter<List<Decimal>, dynamic> get decoder =>
      AnonymousConverter<List<Decimal>, dynamic>((List<Decimal> j) {
        return j.map((f) {
          if (f != null) {
            if (f.isInteger) {
              return f.toString() + ".0";
            } else {
              return f.toString();
            }
          } else {
            return null;
          }
        }).toList();
      });

  @override
  Converter<dynamic, List<Decimal>> get encoder =>
      AnonymousConverter<dynamic, List<Decimal>>((dynamic a) =>
          (a as Iterable).map((dynamic a) =>
              Decimal.tryParse(a as String ?? "-")).toList());
}