import 'dart:convert';
import 'package:bird/smart_json.dart';
import 'package:decimal/decimal.dart';

class SmartTFrom<T extends SmartJSON<T>> extends JSONSmartVariableInterface<T> {

  final T Function(dynamic) fromJSON;

  SmartTFrom(T init)
      : assert(init != null),
        this.fromJSON = init.from,
        super(init)
  ;

  @override
  Codec<dynamic, T> codec() {
    return IsJSONSmartCodec(fromJSON: fromJSON);
  }
}

class SmartTRawFrom<T> extends JSONSmartVariableInterface<T> {

  final T Function(dynamic) fromJSON;

  SmartTRawFrom(T init, this.fromJSON) : super(init);

  @override
  Codec<dynamic, T> codec() {
    return IsJSONSmartCodec(fromJSON: fromJSON);
  }
}

class SmartT<T> extends JSONSmartVariableInterface<T> {

  SmartT(T init) : super(init);

  @override
  Codec<dynamic, T> codec() => PassThroughCodec<T>();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is SmartT &&
              option == other.option &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => option.hashCode;
}

class SmartDynamic extends SmartT<dynamic> {
  SmartDynamic(dynamic init) : super(init);
}

class SmartBool extends SmartT<bool> {

  final bool init;

  SmartBool(this.init) : super(init);

  /// Toggles the bool
  ///
  /// true => false
  /// false => true
  /// null => !init
  void applyToggle() => safeApplySet((a) => a != null ? !a : !init);
}

class SmartInt extends SmartT<int> {

  final int init;

  SmartInt(this.init) : super(init);

  /// Increments the value.
  ///
  /// 5 => 6
  /// 6 => 7
  /// null => init
  void applyIncrement() => safeApplySet((a) => a != null ? a + 1 : init);

  /// Sets the value to the current timestamp in millisecondsSinceEpoch.
  void applyNow() =>
      safeSet(DateTime
          .now()
          .millisecondsSinceEpoch);

  /// Sets the value to the current timestamp in millisecondsSinceEpoch
  /// only if this value is null or -1.
  void applyNowIfNot() {
    if (nullVal() == null || nullVal() == -1) {
      safeSet(DateTime
          .now()
          .millisecondsSinceEpoch);
    }
  }
}

class SmartDouble extends SmartT<double> {
  SmartDouble(double init) : super(init);
}

class SmartString extends SmartT<String> {
  SmartString(String init) : super(init);

  bool get isEmptyOrNull => nullVal() == null || nullVal().isEmpty;
}

/// Lazily initializes [decimal] once on it's first use
/// and resets it when the source string is changed.
///
/// Useful for collections that need decimal values but should
/// not compute them immediately because of performance reasons.
///
/// You may decide for this to not be lazy in which case
/// you should set [isLazy] to false.
///
/// This also accepts not only String values but others
/// like [double] which will be converted to a string
/// with their toString method.
class SmartLazyDecimal extends SmartT<String> {

  final bool isLazy;

  SmartLazyDecimal(String init, {this.isLazy = true}) : super(init);

  Decimal _decimal;

  Decimal get decimal => _decimal ?? _initDecimal();

  @override
  void safeSet(String t) {
    if (isLazy) {
      _decimal = null;
    }

    super.safeSet(t);

    if (!isLazy) {
      _initDecimal();
    }
  }

  @override
  Codec<dynamic, String> codec() =>
      ToStringCodec<String>(fromJSON: (dynamic a) => a?.toString());

  Decimal _initDecimal() =>
      _decimal = option.fold(() => null, Decimal.tryParse);
}

/// Anytime this value is about to be set to null, the initialization value
/// is set instead.
class SmartNonNullableEnum<T> extends SmartTRawFrom<T> {

  final T initT;
  final List<T> fullListOfT;

  SmartNonNullableEnum(this.initT, this.fullListOfT)
      : assert(initT != null),
        assert(fullListOfT != null),
        super(initT, (dynamic a) => firstOfList<T>(fullListOfT, initT)(a));

  bool isThis(T t) => nullVal() == t;

  /// Make sure no null can be set.
  @override
  void safeSet(T t) => super.safeSet(t ?? initT);

  @override
  Codec<dynamic, T> codec() => ToStringCodec(fromJSON: fromJSON);
}
