import 'dart:collection';

import 'package:bird/signal_wave.dart';
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

/// Keeps track of executed commands.
///
/// Inspired by the typical Photoshop Undo/Redo mechanism.
///
/// Command is generic because the user might want to expand
/// commands with for example a description and not have to
/// cast it to the correct command.
abstract class UndoManager<T extends UndoCommand> implements Sink<T> {
  void undo();

  void redo();

  Wave<Option<UndoCommandContainer<T>>> get commands;

  Wave<Option<T>> get nextUndoCommand;

  Wave<Option<T>> get nextRedoCommand;

  Wave<bool> get canUndo;

  Wave<bool> get canRedo;
}

abstract class UndoCommand {
  void execute();

  void undo();
}

class UndoManagerImpl<T extends UndoCommand> extends UndoManager<T> {
  @visibleForTesting
  final Signal<Queue<UndoCommandContainer<T>>> undoList = Signal(Queue.of([]));

  @visibleForTesting
  final Signal<Queue<UndoCommandContainer<T>>> redoList = Signal(Queue.of([]));

  final int undoLimit;

  final int redoLimit;

  @visibleForTesting
  Signal<Option<UndoCommandContainer<T>>> pcommands = Signal(none());

  @override
  Wave<Option<UndoCommandContainer<T>>> get commands => pcommands.wave;

  UndoManagerImpl({
    this.undoLimit = 100,
    this.redoLimit = 100,
    void Function(UndoManagerImpl<T>) onInit,
  }) {
    onInit?.call(this);
  }

  @override
  void add(T cmd) {
    cmd.execute();
    final container =
        UndoCommandContainer(cmd, DRUCommandType.execute, DateTime.now());
    if (canRedo.value) {
      redoList.add(Queue.of([]));
    }
    undoList.add(_removeOverflowFromTheLeft(
        undoList.value..addLast(container), undoLimit));
    pcommands.add(Some(container));
  }

  @override
  void undo() {
    if (canUndo.value) {
      final undos = undoList.value;
      final last = undos.removeLast();
      last.command.undo();
      undoList.add(undos);

      final container = UndoCommandContainer(
          last.command, DRUCommandType.undo, DateTime.now());
      redoList.add(_removeOverflowFromTheRight(
          redoList.value..addLast(container), redoLimit));
      pcommands.add(Some(container));
    }
  }

  @override
  void redo() {
    if (canRedo.value) {
      final redos = redoList.value;
      final last = redos.removeLast();
      last.command.execute();
      redoList.add(redos);

      final container = UndoCommandContainer(
          last.command, DRUCommandType.execute, DateTime.now());
      undoList.add(_removeOverflowFromTheLeft(
          undoList.value..addLast(container), undoLimit));
      pcommands.add(Some(container));
    }
  }

  Queue<UndoCommandContainer<T>> _removeOverflowFromTheLeft(
      Queue<UndoCommandContainer<T>> t, int max) {
    while (t.length > max) {
      t.removeFirst();
    }
    return t;
  }

  Queue<UndoCommandContainer<T>> _removeOverflowFromTheRight(
      Queue<UndoCommandContainer<T>> t, int max) {
    while (t.length > max) {
      t.removeLast();
    }
    return t;
  }

  @override
  Wave<bool> get canRedo =>
      redoList.wave.map((a) => a.isNotEmpty);

  @override
  Wave<bool> get canUndo =>
      undoList.wave.map((a) => a.isNotEmpty);

  @override
  void close() {
    undoList.close();
    redoList.close();
    pcommands.close();
  }

  @override
  Wave<Option<T>> get nextUndoCommand {
    return undoList.wave.map((a) {
      if (a.isNotEmpty) return Some<T>(a.last.command);
      return None<T>();
    });
  }

  @override
  Wave<Option<T>> get nextRedoCommand {
    return redoList.wave.map((a) {
      if (a.isNotEmpty) return Some<T>(a.last.command);
      return None<T>();
    });
  }
}

enum DRUCommandType { execute, undo }

/// Added == and hashcode for testing purposes. They ignore [executedAt].
class UndoCommandContainer<T extends UndoCommand> {
  final T command;
  final DRUCommandType commandType;
  final DateTime executedAt;

  const UndoCommandContainer(this.command, this.commandType, this.executedAt);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is UndoCommandContainer &&
          runtimeType == other.runtimeType &&
          commandType == other.commandType &&
          command == other.command;

  @override
  int get hashCode => command.hashCode ^ commandType.hashCode;
}
