import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

/// Error objects for describing errors.
///
/// All errors inherit [AllErrors]. These errors describe problems that
/// arise from handling local data and working with APIs.
///
/// Consider creating more errors if a common error was not implemented
/// and if it could help the user resolve a problem they are having.
/// If you add more errors then don't forget to also add that error to the
/// mapping function of the super class.
///
/// # [StorageErrors]
/// These errors describe errors that can happen when data is accessed locally.
///
/// - [FailedParsingStorageError] for when e.g. JSON data is loaded from disk
///    and needs to be parsed but it couldn't (if the JSON was invalid for example).
/// - [DataNotFoundStorageError] for when data was expected to be present somewhere but does
///    not exist at the expected location.
/// - [UnknownStorageError] If the above errors do not describe the error then use this one.
///
///
/// # [APIErrors]
/// These errors describe errors that arise when dealing with APIs from an external location.
///
/// - [CancelledAPIError] The request was cancelled by the user.
/// - [LocalNotEnoughCreditsAvailableAPIError] APIs may have a limit. Different policies
///    can limit the usage of certain methods. If it is calculated locally that an API
///    path has run out of credits then this should be used.
/// - [NoInternetAPIError] There is no certain way to determine if an API call was cancelled
///    because there was no internet but a somewhat sure way to do that is to check for a
///    "No host found" message inside the error signature. Use this error when doing that.
/// - [APIStatusError] Non 2XX status messages should be put here.
/// - [TimeoutAPIError] For when request times out.
/// - [APIResponseCustomErrorDetectedAPIError] Some response data contain custom error
///    messages that are specified by the vendor. These errors should be put here.
/// - [UnknownAPIError] If none of the above errors describe the error then use this one.
///
abstract class AllErrors {
  const AllErrors._();

  T map<T>({
    @required T Function(APIErrors) api,
    @required T Function(StorageErrors) storage,
  }) {
    final self = this;
    if (self is APIErrors) {
      return api(self);
    }
    if (self is StorageErrors) {
      return storage(self);
    }
    throw Exception("Invalid");
  }
}

/// Errors than can occur while loading data from an API.
abstract class APIErrors extends AllErrors {
  const APIErrors._() : super._();

  T mapAPIError<T>({
    @required T Function(LocalNotEnoughCreditsAvailableAPIError) credits,
    @required T Function(NoInternetAPIError) noInternet,
    @required T Function(UnknownAPIError) unknown,
    @required T Function(TimeoutAPIError) timeout,
    @required T Function(CancelledAPIError) cancelled,
    @required T Function(APIResponseCustomErrorDetectedAPIError) customError,
    @required T Function(APIStatusError) status,
    @required T Function(APIUnexpectedResponseError) unexpectedResponse,
  }) {
    final self = this;
    if (self is LocalNotEnoughCreditsAvailableAPIError) {
      return credits(self);
    }
    if (self is NoInternetAPIError) {
      return noInternet(self);
    }
    if (self is UnknownAPIError) {
      return unknown(self);
    }
    if (self is TimeoutAPIError) {
      return timeout(self);
    }
    if (self is CancelledAPIError) {
      return cancelled(self);
    }
    if (self is APIResponseCustomErrorDetectedAPIError) {
      return customError(self);
    }
    if (self is APIStatusError) {
      return status(self);
    }
    if (self is APIUnexpectedResponseError) {
      return unexpectedResponse(self);
    }
    throw Exception("Invalid");
  }

  // ignore: use_to_and_as_if_applicable
  Either<APIErrors, T> left<T>() => Left(this);
}

const UnknownAPIError unknownAPIError = UnknownAPIError(None(), None());

UnknownAPIError unknownHintedAPIError(String description,
    {StackTrace stacktrace}) =>
    UnknownAPIError(optionOf(description), optionOf(stacktrace));

/// It is not known why the error occurred.
///
/// Neither [description] nor [stacktrace] go into the identity of this object.
class UnknownAPIError extends APIErrors {
  /// Most of the time Exception.toString().
  final Option<String> description;

  /// The stacktrace for the error.
  final Option<StackTrace> stacktrace;

  const UnknownAPIError(this.description, this.stacktrace) : super._();

  @override
  String toString() =>
      'UnknownAPIError{description: $description, stacktrace: $stacktrace';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is UnknownAPIError && runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

const APIResponseCustomErrorDetectedAPIError vendorError =
APIResponseCustomErrorDetectedAPIError();

/// ------- API Response
///
/// API responses can have error codes and error messages
/// if an api provides custom error messages and one of such
/// messages were found then use this error.
class APIResponseCustomErrorDetectedAPIError extends APIErrors {
  const APIResponseCustomErrorDetectedAPIError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is APIResponseCustomErrorDetectedAPIError &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

APIUnexpectedResponseError unexpectedResponseError(
    ResponseInformation response) =>
    APIUnexpectedResponseError(response);

/// When the response did not meet a criteria then this error should be thrown.
///
/// E.g. when the content type did not match up but should have this can be returned.
///
/// Be careful for == and hashcode only the type counts.
class APIUnexpectedResponseError extends APIErrors {
  final ResponseInformation response;

  const APIUnexpectedResponseError(this.response) : super._();

  @override
  String toString() => 'APIUnexpectedResponseError{response: $response}';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is APIUnexpectedResponseError && runtimeType == other
              .runtimeType;

  @override
  int get hashCode => 0;
}

class ResponseInformation {
  final Option<int> status;
  final Option<String> data;
  final Option<List<String>> contentType;

  const ResponseInformation({
    @required this.status,
    @required this.data,
    @required this.contentType,
  });

  const ResponseInformation.none()
      : status = const None(),
        data = const None(),
        contentType = const None();

  @override
  String toString() => 'ReponseInformation{'
      'status: ${status.map((a) => a.toString()).getOrElse(() => "/")}, '
      'data: ${data.getOrElse(() => "/")}, '
      'contentType: ${contentType.getOrElse(() => ["/"])}}';

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is ResponseInformation &&
              runtimeType == other.runtimeType &&
              status == other.status &&
              data == other.data &&
              contentType == other.contentType;

  @override
  int get hashCode => status.hashCode ^ data.hashCode ^ contentType.hashCode;
}

const TimeoutAPIError timeoutError = TimeoutAPIError();

/// If a running request has timed out pass this error.
class TimeoutAPIError extends APIErrors {
  const TimeoutAPIError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is TimeoutAPIError;

  @override
  int get hashCode => 0;
}

APIStatusError statusError(int status, [String body]) =>
    APIStatusError(optionOf(status), optionOf(body));

/// This is for status errors e.g. 400, 413, ...
class APIStatusError extends APIErrors {
  final Option<int> statusCode;

  final Option<String> body;

  APIStatusError(this.statusCode, this.body) : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is APIStatusError &&
              runtimeType == other.runtimeType &&
              statusCode == other.statusCode;

  @override
  int get hashCode => statusCode.hashCode;

  @override
  String toString() => '$APIStatusError{statusCode: $statusCode, body: $body}';
}

const NoInternetAPIError noInternetError = NoInternetAPIError();

/// ------- API Request
///
/// "No host found" message that indicates
/// no internet connection available.
class NoInternetAPIError extends APIErrors {
  const NoInternetAPIError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is NoInternetAPIError;

  @override
  int get hashCode => 0;
}

const LocalNotEnoughCreditsAvailableAPIError noCreditsError =
LocalNotEnoughCreditsAvailableAPIError();

/// ------- Credits
///
/// Pass this when there are not enough credits found in the credit manager to execute the call
class LocalNotEnoughCreditsAvailableAPIError extends APIErrors {
  const LocalNotEnoughCreditsAvailableAPIError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is LocalNotEnoughCreditsAvailableAPIError &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

const CancelledAPIError cancelledError = CancelledAPIError();

/// If a request was either cancelled by the user or the server
class CancelledAPIError extends APIErrors {
  const CancelledAPIError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is CancelledAPIError;

  @override
  int get hashCode => 0;
}

/// Errors than can occur while loading data from local storage.
abstract class StorageErrors extends AllErrors {
  const StorageErrors._() : super._();

  T mapStorageError<T>({
    @required T Function(DataNotFoundStorageError) dataNotFound,
    @required T Function(FailedParsingStorageError) parsing,
    @required T Function(UnknownStorageError) unknown,
  }) {
    final self = this;
    if (self is DataNotFoundStorageError) {
      return dataNotFound(self);
    }
    if (self is FailedParsingStorageError) {
      return parsing(self);
    }
    if (self is UnknownStorageError) {
      return unknown(self);
    }
    throw Exception("Invalid");
  }

  Either<StorageErrors, T> asLeft<T>() => Left(this);
}

const UnknownStorageError unknownStorageError =
UnknownStorageError(None(), None());

UnknownStorageError unknownHintedStorageError(String description,
    {StackTrace stacktrace}) =>
    UnknownStorageError(optionOf(description), optionOf(stacktrace));

/// General unknown error.
///
/// For == and hashCode only the type counts.
class UnknownStorageError extends StorageErrors {
  /// Usually exception.toString().
  final Option<String> description;

  final Option<StackTrace> stacktrace;

  const UnknownStorageError(this.description, this.stacktrace) : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is UnknownStorageError && runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

const DataNotFoundStorageError notFoundInStorageError =
DataNotFoundStorageError();

class DataNotFoundStorageError extends StorageErrors {
  const DataNotFoundStorageError() : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is DataNotFoundStorageError && runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

FailedParsingStorageError failedParsingError(String desc) =>
    FailedParsingStorageError(desc);

class FailedParsingStorageError extends StorageErrors {
  final String description;

  const FailedParsingStorageError(this.description) : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is FailedParsingStorageError && runtimeType == other
              .runtimeType;

  @override
  int get hashCode => 0;
}
