import 'dart:async';
import 'dart:convert';

import 'package:bird/dart_filter.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:crypto/crypto.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:meta/meta.dart';

abstract class WebRequestService {
  WebRequestInterface get requestInterface;
}

abstract class WebRequestInterface {
  Future<Either<APIErrors, String>> request(APIRestMethodDescription method);
}

/// Rest API Definitions should extend this.
/// Inherit and use [get], [getTyped], [post] and [postTyped] to create endpoints.
class APIRestDescription {
  /// With no trailing slashes. Is assumed to be https.
  final String host;

  /// No leading or trailing slashes. If set to null then this will be ignored.
  final String path;

  const APIRestDescription({
    @required this.host,
    this.path,
  });

  String get getNonEmptyPath => path == null || path.isEmpty ? null : path;

  /// Subclasses can call this to define methods.
  ///
  /// Convenience method.
  APIRestMethodDescription get({
    Iterable<FilterCriteria<String>> contentType,
    String path,
    String cmd,
    Map<String, Object /*String|Iterable<String>*/ > Function(
            APIRestMethodDescription)
        args,
    Map<String, String> Function(APIRestMethodDescription) headers,
  }) =>
      APIRestMethodDescription(
        api: this,
        path: path,
        cmd: cmd,
        expectedContentType: contentType ?? [],
        arguments: APIHTTPMethod.get(
          headers: headers ?? _emptyHeadersInit,
          getParam: args,
        ),
      );

  /// Subclasses can call this to define methods.
  ///
  /// Convenience method.
  APIRestMethodDescription post({
    Iterable<FilterCriteria<String>> contentType,
    String path,
    String cmd,
    String Function(APIRestMethodDescription, BodyHelper) body,
    Map<String, String> Function(APIRestMethodDescription) headers,
  }) =>
      APIRestMethodDescription(
        api: this,
        path: path,
        cmd: cmd,
        expectedContentType: contentType ?? [],
        arguments: APIHTTPMethod.post(
          headers: headers ?? _emptyHeadersInit,
          postBody: body,
        ),
      );
}

Map<String, String> _emptyHeadersInit(APIRestMethodDescription a) => {};

/// Describes an http method endpoint.
///
/// REST-API Layer 2: API Method Definition.
///
/// These should be created inside a subclass of [APIRestDescription].
/// Subclasses of [APIRestDescription] should call the convenient [get],
/// [post] or others to provide new endpoints.
class APIRestMethodDescription {
  /// Precedure to access remote call.
  final APIHTTPMethod arguments;

  /// API for description.
  final APIRestDescription api;

  /// API method path.
  final String path;

  /// Same as [path] but used not in URI but in signatures or the post body.
  final String cmd;

  final String scheme;

  /// Used inside Layer 4 to store responses of methods.
  String get uniqueIdentifier => [
        api.host,
        arguments.httpMethod,
        api.path,
        path,
        cmd,

        /// Hash the query because that value could become too large
        /// to store on the device if used inside a folder name
        /// because folder names need to be of a certain length.
        if (arguments.getParam != null)
          md5.convert(utf8.encode(buildUri().query)),
        if (arguments.postBody != null)
          md5.convert(
              utf8.encode(arguments.postBody(this, const BodyHelper()))),
      ].where((a) => a != null).join("/");

  /// Set the expected content type.
  ///
  /// If nothing is set then every contentType is assumed to be valid.
  ///
  /// If something is set then only the expected content type inside the
  /// response will be let through without an error. The response of
  /// content types that do not match up will be wrapped in an error.
  ///
  /// Useful for passing the user HTML that some pages will show when
  /// an API decides to go into maintenance mode, DDOS protection kicks in or
  /// other reasons. Some of these errors can be handled (e.g. when a captcha
  /// needs to be filled out).
  ///
  /// That's why its important to limit the content type for requests.
  /// JSON Requests most of the time start with "application/json"
  /// and some contain a charset which would look like this: "application/json; charset=utf-8".
  ///
  /// Take a look at [APIContentTypes] for common predefined content type criteria.
  final Iterable<FilterCriteria<String>> expectedContentType;

  const APIRestMethodDescription({
    @required this.api,
    @required this.arguments,
    this.path,
    this.cmd,
    this.scheme = "https",
    this.expectedContentType = const [],
  })  : assert(api != null),
        assert(arguments != null),
        assert(expectedContentType != null),
        assert(scheme != null);

  Future<DataResponse<String>> methodFetch(WebRequestService requestService) =>
      arguments.lowFetch(this, requestService);

  Uri buildUri() {
    return Uri(
      scheme: scheme,
      host: api.host,

      /// Filter out nulls and add a leading slash in front of the rest.
      path: [api.getNonEmptyPath, path]
          .where((a) => a != null)
          .map((s) => "/$s")
          .join(""),
      queryParameters: arguments.getParam
          ?.call(this)
          ?.map<String, dynamic>((key, dynamic value) {
        dynamic val;

        if (value is String || value is Iterable<String>) {
          val = value;
        } else if (value is int) {
          val = value.toString();
        } else {
          throw ArgumentError(
              "Url query parameters can only consist of values that are either a String, an integer or a list of strings.");
        }

        return MapEntry<String, dynamic>(key, val);
      }),
    );
  }
}

/// Describes common content type filter criteria.
/// Used in [APIRestMethodDescription].
class APIContentTypes {
  /// Starts with application/json and ignores the charset.
  static const FilterCriteria<String> anyJson =
      StartsWithFilter("application/json", true);

  /// Starts with application/rss+xml and ignores the charset.
  static const FilterCriteria<String> anyRssXml =
      StartsWithFilter("application/rss+xml", true);

  /// Starts with text/xml and ignores the charset.
  static const FilterCriteria<String> textXml =
      StartsWithFilter("text/xml", true);
}

/// An Http method has many parameters which need to be set.
/// An [APIHTTPMethod] abstracts these. These are passed to created
/// instances of [APIRestMethodDescription].
class APIHTTPMethod {
  /// HTTP urls can have url encoded parameters which look like this
  /// ?123=xyz&abc=xyz
  final Map<String, Object /*String|Iterable<String>*/ > Function(
      APIRestMethodDescription) getParam;

  /// HTTP Headers can be set here.
  final Map<String, String> Function(APIRestMethodDescription) headers;

  /// an [APIHTTPMethod] with a "POST" httMethod can also have a
  /// postBody besides having parameters at [getParam].
  final String Function(APIRestMethodDescription, BodyHelper) postBody;

  /// The HttpMethod as defined by the HTTP standard.
  final String httpMethod;

  /// Use [APIRestDescription.get] to define methods.
  const APIHTTPMethod.get({
    this.headers = _emptyHeadersInit,
    this.getParam,
  })  : assert(headers != null),
        postBody = null,
        httpMethod = "GET";

  /// Use [APIRestDescription.post] to define methods.
  const APIHTTPMethod.post({
    this.headers = _emptyHeadersInit,
    this.getParam,
    this.postBody,
  })  : assert(headers != null),
        httpMethod = "POST";

  Future<DataResponse<String>> lowFetch(
      APIRestMethodDescription method, WebRequestService service) async {
    return DataResponse.measure<String>(
        () => service.requestInterface.request(method));
  }
}

/// The response that is received by an api call.
class DataResponse<T> {
  /// Tells when the call started.
  final DateTime started;

  /// Tells when the call ended.
  final DateTime ended;

  /// An error or the response of a request.
  final Either<APIErrors, T> response;

  Duration get took => ended.difference(started);

  static Future<DataResponse<T>> measure<T>(
      Future<Either<APIErrors, T>> Function() data) async {
    final started = DateTime.now();
    final response = await data();
    return DataResponse<T>(
      started: started,
      response: response,
      ended: DateTime.now(),
    );
  }

  DataResponse({
    @required this.started,
    @required this.ended,
    @required this.response,
  })  : assert(started != null),
        assert(ended != null),
        assert(response != null),
        assert(started.millisecondsSinceEpoch <= ended.millisecondsSinceEpoch);

  @override
  String toString() => '$DataResponse{'
      '\n\tstarted: $started, '
      '\n\tended: $ended, '
      '\n\tresponse: $response'
      '\n}';
}

/// REST-API Layer 3: Model definitions and updates.
///
/// Models are defined as [SmartJSON] objects.
/// Every model that can get updates from an API endpoint
/// should provide a [ModelUpdate] object:
///
///     ModelUpdate<...> update = ModelUpdate(...);
///
/// This abstracts layer 1 & 2 by giving all models
/// that can update themselves a way to do so.
///
/// Because Models can't always update themselves by one call,
/// different implementations of [ModelUpdate] for different
/// use cases exist:
///
/// - [OneModelUpdate] depends on a single endpoint.
abstract class RAWModelUpdate<T, RAW> {
  /// Only gets data from it's source and converts it
  /// with no other local side effects.
  /// Caching, storage and such are done in next layers.
  ///
  /// You can optionally provide a [CancelToken]
  /// to be able to cancel requests.
  /// TODO change that. this is bad. requestOnlyT should be enough. implementations should specify everything else
  Future<Either<AllErrors, Tuple2<T, RAW>>> request();

  Future<Either<AllErrors, T>> requestOnlyT() async =>
      request().then((a) => a.fold(left, (a) => right(a.value1)));

  const RAWModelUpdate();
}

abstract class ModelUpdate<T> extends RAWModelUpdate<T, String> {
  const ModelUpdate();
}

/// Layer 3
///
/// For models that can get all their data from one endpoint.
class OneModelUpdate<T extends SmartJSON<T>> extends ModelUpdate<T> {
  final WebRequestService requestService;

  final APIRestMethodDescription method;

  final T Function() tCreator;

  const OneModelUpdate({
    @required this.method,
    @required this.tCreator,
    @required this.requestService,
  })  : assert(method != null),
        assert(tCreator != null),
        assert(requestService != null);

  factory OneModelUpdate.short(WebRequestService requestService,
      APIRestMethodDescription method, T Function() creator) {
    return OneModelUpdate(
      method: method,
      tCreator: creator,
      requestService: requestService,
    );
  }

  @override
  Future<Either<APIErrors, Tuple2<T, String>>> request() async => method
          .methodFetch(requestService)
          .then((DataResponse<String> dataResponse) {
        return dataResponse.response.fold(left, (String a) {
          return _catching(() {
            return Tuple2(tCreator().stringCodec().decode(a), a);
          }).leftMap<APIErrors>((MapEntry<dynamic, StackTrace> error) {
            return unknownHintedAPIError(error.key.toString(),
                stacktrace: error.value);
          });
        });
      });

  Either<MapEntry<dynamic, StackTrace>, A> _catching<A>(Function0<A> thunk) {
    try {
      return right(thunk());
    } catch (e, s) {
      return left(MapEntry<dynamic, StackTrace>(e, s));
    }
  }
}

class RawOneModelUpdate<T> extends ModelUpdate<T> {
  final APIRestMethodDescription method;

  final T Function(String) tCreator;

  final WebRequestService requestService;

  const RawOneModelUpdate({
    @required this.method,
    @required this.tCreator,
    @required this.requestService,
  })  : assert(method != null),
        assert(tCreator != null),
        assert(requestService != null);

  factory RawOneModelUpdate.short(WebRequestService requestService,
      APIRestMethodDescription method, T Function(String) tCreator) {
    return RawOneModelUpdate(
        method: method, tCreator: tCreator, requestService: requestService);
  }

  @override
  Future<Either<APIErrors, Tuple2<T, String>>> request() async => method
      .methodFetch(requestService)
      .then((a) => a.response.map((a) => Tuple2(tCreator(a), a)));
}

/// API Layer 4 version of [OneModelUpdate].
///
/// Example:
/// - Get some static data like the current status of a service.
/// - Get a ticker from one endpoint that doesn't require anything else.
/// - Coinbase ticker.
class APISingleDataAccess<T extends SmartJSON<T>>
    extends RAWModelUpdate<T, String> {
  final OneModelUpdate<T> modelUpdate;

  const APISingleDataAccess({
    @required this.modelUpdate,
  }) : assert(modelUpdate != null, #b);

  @override
  Future<Either<AllErrors, Tuple2<T, String>>> request() async {
    final a = await modelUpdate.request();

    /// Map new data:
    return await a.fold(left, (r) async {
      /// Return parsed model:
      return right(r);
    });
  }
}

/// API Layer 4.
///
/// Used for 2 [APISingleDataAccess] where one depends on the other.
///
/// Example:
/// - When you need id's to fetch tickers => tickers depend on id's.
class APIDoubleDataAccess<T, U, USOURCE> extends RAWModelUpdate<U, USOURCE> {
  final RAWModelUpdate<T, String> _freeModelUpdate;

  final RAWModelUpdate<U, USOURCE> Function(T) _dependentModelUpdate;

  const APIDoubleDataAccess({
    @required RAWModelUpdate<T, String> freeModelUpdate,
    @required RAWModelUpdate<U, USOURCE> Function(T) dependentModelUpdate,
  })  : assert(freeModelUpdate != null, #a),
        assert(dependentModelUpdate != null, #b),
        this._freeModelUpdate = freeModelUpdate,
        this._dependentModelUpdate = dependentModelUpdate;

  @override
  Future<Either<AllErrors, Tuple2<U, USOURCE>>> request() {
    /// Load independent data first:
    final fetch = _freeModelUpdate.requestOnlyT();
    assert(fetch != null, #a);

    return fetch.then((independent) async {
      assert(independent != null, #b);

      /// Use new independent data to fetch dependent data:
      return await independent.fold(left, (newData) {
        /// Fetch data from the dependent model:
        final dependent = _dependentModelUpdate(newData);
        assert(dependent != null, #c);

        /// Test #d.
        return dependent.request();
      });
    });
  }
}

/// API Layer 4.
///
/// For models that can get all it's members by getting data from one endpoint
/// and passing that iteratively to a second one which responses will be accumulated.
///
/// Example:
/// - To get tickers you need to fetch valid id's. But you can only
///   get a single ticker for one id therefore you need to call one
///   endpoint multiple times.
/// - Bitstamp ticker.
class APIAccumulatedDataAccess<U, G extends SmartJSON<G>>
    extends RAWModelUpdate<IList<G>, IList<String>> {
  final Iterable<U> _source;

  final APISingleDataAccess<G> Function(U) _modelToUpdate;

  const APIAccumulatedDataAccess({
    @required Iterable<U> source,
    @required APISingleDataAccess<G> Function(U) modelToUpdate,
  })  : assert(source != null, #a),
        assert(modelToUpdate != null, #b),
        this._source = source,
        this._modelToUpdate = modelToUpdate;

  @override
  Future<Either<AllErrors, Tuple2<IList<G>, IList<String>>>> request() async {
    final List<Either<AllErrors, Tuple2<G, String>>> newData = [];

    /// Keep track of a single error and don't load any other data if one error is present:
    Either<AllErrors, Tuple2<IList<G>, IList<String>>> stopOnSingleError;

    await Future.forEach(_source.map(_modelToUpdate),
        (APISingleDataAccess<G> single) async {
      /// Continue loading if no error is present:
      if (stopOnSingleError == null) {
        assert(single != null, #a);
        final Either<AllErrors, Tuple2<G, String>> data =
            await single.request();
        assert(data != null, #b);
        data.fold((error) {
          /// An error occurred while loading data, set to stop all further loading:
          stopOnSingleError = left(error);
        }, (data) {
          /// New data received successfully, accumulate.
          newData.add(right(data));
        });
      }
    });

    if (stopOnSingleError != null) {
      /// If there was an error then return that:
      /// Test #c
      /// Test #e
      return stopOnSingleError;
    } else {
      /// Create a list from the loaded data and take the either out:
      /// Test #d
      final fullList = IList.sequenceEither(IList.from(newData));
      return fullList.fold(
          left,
          (a) => right(Tuple2(
              a.map((tuple) => tuple.value1), a.map((tuple) => tuple.value2))));
    }
  }
}
