import 'package:bird/signal_wave.dart';
import 'package:bird/signal_wave_plus.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:pedantic/pedantic.dart';

/// [ServiceEndpoint] implementation for REST endpoints which are defined through [RAWModelUpdate] implementations.
class RestServiceEndpoint<T> extends ServiceEndpoint<T> {
  /// Tells how to get this data.
  final RAWModelUpdate<T, Object> _model;

  /// Tells in which time periods the [observeLast] observable should try to fetch new data.
  final Duration fetchFrequency;

  bool _isRunning = false;
  Signal<Option<Either<AllErrors, T>>> _lastData;
  Signal<APIState> _lastState;

  RestServiceEndpoint(this._model,
      {this.fetchFrequency = const Duration(milliseconds: 2000)}) {
    _lastData = Signal(none(), onListen: restartObservable, onCancel: () {
      if (_lastState.value is! APIStateInitialized &&
          _lastState.value is! APIStateDisconnected) {
        _lastState.add(apiInitialized);
      }
    });
    _lastState = Signal(apiInitialized);
  }

  Future<void> restartObservable() async {
    if (!_isRunning) {
      _isRunning = true;

      if (lastState.value is! APIStateConnected) _safeChangeState(apiLoading);

      final result = await singleLastData;

      await result.fold(
        (a) {
          _isRunning = false;
          _safeChangeState(apiDisconnected(a));
          if (!_lastData.isClosed) _lastData.add(Some(result));
        },
        (data) async {
          if (lastState.value is! APIStateConnected) {
            _safeChangeState(apiConnected);
          }
          if (!_lastData.isClosed) _lastData.add(Some(result));
          await Future<dynamic>.delayed(fetchFrequency);
          _isRunning = false;
          if (_lastData.hasSubscribers) {
            unawaited(restartObservable());
          } else {
            _safeChangeState(apiInitialized);
          }
        },
      );
    }
  }

  /// Simply loads data from the provided rest data model.
  ///
  /// Base operation for the observable.
  Future<Either<AllErrors, T>> get singleLastData async {
    return (await _model.request()).map((f) => f.value1);
  }

  /// Receives data regularly in [fetchFrequency] intervals as long as there are listeners.
  ///
  /// This being a getter ensures that the synchronously available value is up-to-date.
  @override
  Wave<Option<Either<AllErrors, T>>> get lastData => _lastData.wave;

  /// Provides the state of the [lastData]s underlying logic.
  ///
  /// This being a getter ensures that the synchronously available value is up-to-date.
  @override
  Wave<APIState> get lastState => _lastState.wave;

  /// Will restart [lastData] when the last returned data wasn't an error.
  /// Therefore it oly does something when [lastState] is [APIStateDisconnected].
  @override
  Future<void> reconnect() async {
    if (lastState.value is APIStateDisconnected) {
      _lastState.add(apiInitialized);
      await restartObservable();
    }
  }

  @override
  Future<void> dispose() async {
    if (_lastState.value is! APIStateDisconnected) {
      _lastState.add(apiDisconnected(cancelledError));
    }
    _lastData.close();
    _lastState.close();
  }

  void _safeChangeState(APIState state) {
    if (!_lastState.isClosed) _lastState.add(state);
  }
}

class StaticServiceEndpointImpl<T> implements StaticServiceEndpoint<T> {
  final ServiceEndpoint<T> endpoint;

  Signal<Option<Either<AllErrors, T>>> _lastData;
  Signal<APIState> _lastState;

  /// Setting [autoUpdate] to true will automatically update the data when the first
  /// listener listens to [_lastData].
  StaticServiceEndpointImpl(this.endpoint, bool autoUpdate) {
    _lastData = Signal(endpoint.lastData.value, onListen: () {
      if (autoUpdate) update();
    });
    lastData = _lastData.wave;
    _lastState = Signal(const APIStateInitialized());
    lastState = _lastState.wave;

    error = lastState
        .map((a) => a.hasError)
        .and(lastData.left())
        .latest((a, b) => a.orElse(() => b));

    data = lastData.right();

    loading = lastState.map((a) => a.isLoading);
  }

  Wave<Option<AllErrors>> error;
  Wave<bool> loading;
  Wave<Option<T>> data;

  @override
  Wave<Option<Either<AllErrors, T>>> lastData;

  @override
  Wave<APIState> lastState;

  @override
  Future<void> update() async {
    if (lastState.value != apiLoading && _lastData.hasSubscribers) {
      _safeAddState(apiLoading);

      (await endpoint.lastData.nextAsFuture()).visit((resultData) {
        resultData.fold(() {
          _safeAddState(const APIStateDisconnected(Some(cancelledError)));
          _safeAddData(none());
        }, (data) {
          data.fold((error) {
            _safeAddState(APIStateDisconnected(some(error)));
            _safeAddData(Some(Left(error)));
          }, (d) {
            _safeAddState(apiConnected);
            _safeAddData(Some(Right(d)));
          });
        });
      }, () {
        _safeAddState(const APIStateDisconnected(Some(cancelledError)));
        _safeAddData(none());
      });
    }
  }

  void _safeAddState(APIState state) {
    if (!_lastState.isClosed) _lastState.add(state);
  }

  void _safeAddData(Option<Either<AllErrors, T>> data) {
    if (!_lastData.isClosed) _lastData.add(data);
  }

  @override
  Future<void> dispose() async {
    if (_lastState.value is! APIStateDisconnected) {
      _lastState.add(apiDisconnected(cancelledError));
    }
    _lastData.close();
    _lastState.close();
  }
}

/// TODO replace as soon as static extension methods land.
/// TODO It can then be described as an extension on
/// TODO ServiceEndpointStreams<T> where T extends OptionalParsedResult<T>
class ServiceEndpointConvenienceStreams<T> {
  final ServiceEndpointStreams<OptionalParsedResult<T>> source;

  ServiceEndpointConvenienceStreams(this.source) {
    liftedData = source.lastData.map(OptionalParsedResult.lift);
    relevantData = liftedData.where((a) => a.isSome());
    hasRelevantData = relevantData.isSome();

    getErrors = source.lastState
        .map((a) => a.hasError)
        .and(liftedData.left())
        .latest((a, b) => a.orElse(() => b));

    getData = relevantData.right();
    isLoading = source.lastState.map((a) => a.isLoading);
  }

  /// Lifts the [ParsingError] out of the results.
  Wave<Option<Either<AllErrors, T>>> liftedData;

  /// Lets only Some through
  ///
  /// The danger with this is that the seed which will be [None] can't be filtered out.
  /// Therefore the option has to be mapped a default data for that case.
  Wave<Option<Either<AllErrors, T>>> relevantData;

  Wave<bool> isLoading;

  Wave<R> onlyRelevantEntries<R>(R defaultInit, R onError, R Function(T) map) {
    return relevantData.map((a) {
      return a.fold(() => defaultInit, (b) => b.fold((_) => onError, map));
    });
  }

  Wave<bool> hasRelevantData;

  Wave<Option<AllErrors>> getErrors;

  Wave<Option<T>> getData;
}
