import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:meta/meta.dart';

abstract class SocketEndpointInterface extends ServiceEndpoint<SocketInterface>
    implements ReducerStateInterface {
  // ignore: close_sinks
  final Signal<APIState> _lastState = Signal(apiInitialized);

  // ignore: close_sinks
  final Signal<Option<StringConversion>> _socketData = Signal(none());

  // ignore: close_sinks
  Signal<Option<Either<AllErrors, SocketInterface>>> _lastData;

  @override
  Wave<APIState> get lastState =>
      _lastState.wave; // A getter ensures that .value is always up-to-date.

  @override
  Wave<Option<Either<AllErrors, SocketInterface>>> get lastData =>
      _lastData.wave; // A getter ensures that .value is always up-to-date.

  Wave<Option<StringConversion>> get socketData =>
      _socketData.wave; // A getter ensures that .value is always up-to-date.
}

class SocketEndpointImpl extends SocketEndpointInterface {
  final Future<SocketInterface> Function() socketCreator;
  final IsConnectedPolicy whenIsConnectedPolicy;

  bool _isDisposed = false;

  SocketEndpointImpl({
    @required this.socketCreator,
    @required this.whenIsConnectedPolicy,
    this.reducers = const [],
  }) {
    _lastData = Signal(
      none(),
      onListen: onConnected,
      onCancel: () {
        final newValue = addInitWhenActive(_lastState.value);
        if (newValue != null) _lastState.add(newValue);
        _currentSubscription?.cancel();
      },
    );
    reducers.forEach((a) => a.register(this));
  }

  final List<SocketChannelReducer> reducers;

  Disposable _currentSubscription;

  Future<SocketInterface> _initSocket(Signal<APIState> state) async {
    if (state.value is! APIStateInitialized) state.add(apiInitialized);
    state.add(apiLoading);
    return socketCreator();
  }

  Future<void> onConnected() {
    return _initSocket(_lastState).then((socket) {
      if (_isDisposed) {
        socket.close();
        return;
      }

      _lastData.add(Some(Right(socket)));

      _currentSubscription = socket.stream.subscribe((contents) {
        if (whenIsConnectedPolicy is ConnectImmediatelyPolicy &&
            _lastState.value is! APIStateConnected) {
          _lastState.add(apiConnected);
        }

        contents.fold(() {}, (content) {
          if (whenIsConnectedPolicy is IsConnectedOnFirstMessagePolicy &&
              _lastState.value is! APIStateConnected) {
            _lastState.add(apiConnected);
          }

          content.fold(disconnect, (dynamic data) {
            if (data is String) {
              StringConversion(data)
                  .apply<dynamic>((a) => _socketData.add(Some(a)))
                  .apply<dynamic>((msg) =>
                  runReducers(
                      reducers, msg, _lastState.value, _lastState.add));
            }
          });
        });
      }, () => socket.close());
    });
  }

  @override
  void sendMessage(String msg) {
    if (lastState.value is APIStateConnected) {
      lastData.value.fold(() => throw Exception("no socket yet"), (val) {
        val.fold((error) {
          throw Exception(
              "Invalid state"); // socket can't be connected and in error
        }, (socket) => socket.add(msg));
      });
    }
  }

  @override
  void reconnect() {
    if (lastState.value is APIStateDisconnected &&
        _lastData.hasSubscribers &&
        !_isDisposed) onConnected();
  }

  @override
  void disconnect(AllErrors error) {
    if (_lastState.value is! APIStateDisconnected) {
      _lastData.add(Some(Left(error)));
      _lastState.add(apiDisconnected(error));
    }
    _currentSubscription?.cancel();
  }

  @override
  Future<void> dispose() async {
    _isDisposed = true;
    _currentSubscription?.cancel();
    reducers.forEach((a) => a.dispose());
    if (_lastState.value is! APIStateDisconnected) {
      _lastState.add(apiDisconnected(cancelledError));
    }
    _lastState.close();
    _lastData.close();
    _socketData.close();
  }
}

APIState addInitWhenActive(APIState value) {
  return value is! APIStateInitialized && value is! APIStateDisconnected
         ? apiInitialized
         : null;
}
