import 'dart:async';

import 'package:bird/base.dart';
import 'package:bird/smart_api.dart';
import 'package:pedantic/pedantic.dart';
import 'package:meta/meta.dart';

/// Every API should have an APIService class that provides [ServiceEndpoint] members
/// for all the data that can be retrieved from that API.
///
/// This class takes care of caching [ServiceEndpoint] calls when the appropriate helper methods
/// like [rest] and [socketSubUnsub] are used. That way you can retrieve different data (of a single type) from an API
/// by defining a single [ServiceEndpoint] call. The passed cacheId should canonicalize the parameters
/// into a single string because that is the lookup key for the various Endpoints and therefore they should also
/// be unique to each ServiceEndpoint of a certain type T.
///
/// For sockets you will need a [SocketEndpointImpl]. If your socket won't be disposed (like by e.g. a channel
/// that owns that socket) then you will need to dispose it in [dispose].
abstract class APIService implements BlocBase {
  /// Hols references to all [ServiceEndpoint]s that were instantiated through [rest] or [socketSubUnsub].
  @visibleForTesting
  final Map<String, ServiceEndpoint> cache = {};

  /// Services wrapped with this method are collected at [cache].
  /// That way you can dispose all of them without specifying them all
  /// in a dispose method.
  @visibleForTesting
  ServiceEndpoint<T> cachedAccess<T>(
      String cacheId, ServiceEndpoint<T> Function() endpoint) {
    cache.putIfAbsent(cacheId, () => endpoint());
    return cache[cacheId] as ServiceEndpoint<T>;
  }

  /// Use this to define [ModelUpdate] rest endpoints.
  ServiceEndpoint<T> rest<T>(String cacheId, RAWModelUpdate<T, Object> model) =>
      cachedAccess(cacheId, () => RestServiceEndpoint<T>(model));

  /// Override this to provide a default socket for [socketForeign] definitions.
  ///
  /// This [SocketEndpointImpl] will be disposed if it is overriden
  @visibleForOverriding
  @visibleForTesting
  SocketEndpointInterface defaultForeignSocket() => null;

  SocketEndpointInterface _getDefaultForeignSocket() {
    if (defaultForeignSocket() != null) {
      return defaultForeignSocket();
    } else {
      throw Exception("If you're using a 'socketForeign' call then please "
          "provide a socket or implement defaultForeignSocket in your definition");
    }
  }

  /// Use this for socket channels.
  ///
  /// Please provide a socket if the service doesn't override [defaultForeignSocket]
  ServiceEndpoint<T> socketForeign<T>({
    @required String cacheId,
    @required SingleChannelSocketDefinition<T> channel,
    SocketEndpointInterface Function() socket,
  }) =>
      cachedAccess(
        cacheId,
        () => SocketChannelServiceEndpoint<T>(
          ownsSocket: false,
          socket: (url) => socket?.call() ?? _getDefaultForeignSocket(),
          channel: channel.preferences,
          converter: channel.thisConverter,
          whenIsConnectedPolicy: channel.preferences.whenIsConnectedPolicy,
        ),
      );

  /// Use this for socket channels where the URL of the socket
  /// connection depends on the channels subscribed to it.
  ServiceEndpoint<T> socketOwn<T>({
    @required String cacheId,
    @required SocketEndpointInterface Function(String url) socketBuilder,
    @required SingleChannelSocketDefinition<T> channel,
  }) =>
      cachedAccess(
        cacheId,
        () => SocketChannelServiceEndpoint<T>(
          ownsSocket: true,
          socket: (provider) => socketBuilder(provider.url),
          channel: channel.preferences,
          channelUrl: channel,
          converter: (str) => channel.converter(str, channel),
          whenIsConnectedPolicy: channel.preferences.whenIsConnectedPolicy,
        ),
      );

  @override
  Future<void> dispose() async {
    if (defaultForeignSocket() != null) {
      unawaited(defaultForeignSocket().dispose());
    }
    cache.values.forEach((a) => a.dispose());
  }
}
