import 'dart:convert';

import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

abstract class JSON_RPC_Constants {
  static const String JSON_RPC_2_0_VERSION_VALUE = "2.0";
  static const String JSONRPC_KEY = "jsonrpc";
  static const String METHOD_KEY = "method";
  static const String PARAMS_KEY = "params";
  static const String ID_KEY = "id";
  static const String RESULT_KEY = "result";
  static const String ERROR_KEY = "error";
}

/// Request object
///
/// Example:
///   {"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": 1}
class JSONRPC2_0RequestObject {
  final String method;

  final Option<Object> params;

  final Option<Object> id;

  JSONRPC2_0RequestObject({
    @required this.method,
    Object id,
    Object params,
  })  : assert(method != null),
        assert(params == null || params is List || params is Map),
        this.id = optionOf(id),
        this.params = optionOf(params);

  factory JSONRPC2_0RequestObject.notification({
    @required String method,
    Object params,
  }) =>
      JSONRPC2_0RequestObject(
        method: method,
        params: params,
      );

  static Either<RPCException, JSONRPC2_0RequestObject> fromJson(
      Map<String, dynamic> map) {
    if (!map.containsKey(JSON_RPC_Constants.JSONRPC_KEY)) {
      return const Left(
          MissingRPCVersionException("JsonRPC version was not found"));
    }

    if (map[JSON_RPC_Constants.JSONRPC_KEY] !=
        JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE) {
      return Left(WrongRPCVersionException(
          "JsonRPC version is not correct, ${map[JSON_RPC_Constants.JSONRPC_KEY]} was found, ${JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE} is expected"));
    }

    if (!map.containsKey(JSON_RPC_Constants.METHOD_KEY)) {
      return const Left(MissingMethodRPCException("Method is missing"));
    }

    final dynamic paramsContent = map[JSON_RPC_Constants.PARAMS_KEY];

    if (paramsContent != null &&
        paramsContent is! List &&
        paramsContent is! Map) {
      return Left(BadParamsRPCException(
          "Params were: $paramsContent with the type ${paramsContent.runtimeType} but can only be a list or a map"));
    }

    return right(JSONRPC2_0RequestObject(
      method: map[JSON_RPC_Constants.METHOD_KEY] as String,
      params: paramsContent,
      id: map[JSON_RPC_Constants.ID_KEY],
    ));
  }

  static Either<RPCException, JSONRPC2_0RequestObject> fromString(String str) {
    try {
      return fromJson(json.decode(str) as Map<String, dynamic>);
    } catch (e) {
      return left(BadDataRPCException(e.toString()));
    }
  }

  Map<String, dynamic> toJson() {
    final result = <String, dynamic>{
      JSON_RPC_Constants.JSONRPC_KEY:
          JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE,
      JSON_RPC_Constants.METHOD_KEY: method,
    };

    id.fold(() {}, (dynamic id) {
      result[JSON_RPC_Constants.ID_KEY] = id;
    });

    params.fold(() {}, (dynamic params) {
      result[JSON_RPC_Constants.PARAMS_KEY] = params;
    });

    return result;
  }

  @override
  String toString() => json.encode(toJson());
}

abstract class RPCException implements Exception {}

/// Data that the object was trying to be parsed from was invalid.
class BadDataRPCException implements RPCException {
  final String cause;

  const BadDataRPCException(this.cause);

  @override
  String toString() => '$BadDataRPCException{cause: $cause}';
}

/// RPC Version was not provided when parsing [JSONRPC2_0RequestObject.fromJson]
class MissingRPCVersionException implements RPCException {
  final String cause;

  const MissingRPCVersionException(this.cause);

  @override
  String toString() => '$MissingRPCVersionException{cause: $cause}';
}

/// RPC Version was incorrect when parsing [JSONRPC2_0RequestObject.fromJson]
class WrongRPCVersionException implements RPCException {
  final String cause;

  const WrongRPCVersionException(this.cause);

  @override
  String toString() => '$WrongRPCVersionException{cause: $cause}';
}

/// Method missing when parsing [JSONRPC2_0RequestObject.fromJson]
class MissingMethodRPCException implements RPCException {
  final String cause;

  const MissingMethodRPCException(this.cause);

  @override
  String toString() => '$MissingMethodRPCException{cause: $cause}';
}

/// When the dynamic params are neither a List nor a Map [JSONRPC2_0RequestObject.fromJson]
class BadParamsRPCException implements RPCException {
  final String cause;

  const BadParamsRPCException(this.cause);

  @override
  String toString() => '$BadParamsRPCException{cause: $cause}';
}

/// Response object
///
/// Example:
///   {"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"}, "id": null}
///   {"jsonrpc": "2.0", "result": 19, "id": 1}
class JSONRPC2_0ResponseObject {
  Option<Object> result;
  Option<JSON_RPC_2_Error> error;
  Object id;

  JSONRPC2_0ResponseObject({
    @required this.result,
    @required this.error,
    @required this.id,
  }) :
        // There must never exist a result and an error at the same time in one object.
        assert(result.isSome() && error.isNone() ||
            result.isNone() && error.isSome());

  static Either<RPCException, JSONRPC2_0ResponseObject> fromJson(
      Map<String, dynamic> map) {
    if (!map.containsKey(JSON_RPC_Constants.JSONRPC_KEY)) {
      return const Left(
          MissingRPCVersionException("JsonRPC version was not found"));
    }

    if (map[JSON_RPC_Constants.JSONRPC_KEY] !=
        JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE) {
      return Left(WrongRPCVersionException(
          "JsonRPC version is not correct, ${map[JSON_RPC_Constants.JSONRPC_KEY]} was found, ${JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE} is expected"));
    }

    if (!map.containsKey(JSON_RPC_Constants.RESULT_KEY) &&
        !map.containsKey(JSON_RPC_Constants.ERROR_KEY)) {
      return const Left(NoContentRPCException(
          "There is neither a result nor an error inside the json data"));
    }

    if (!map.containsKey(JSON_RPC_Constants.ID_KEY)) {
      return const Left(
          NoIDRPCException("Response objects need to have an id"));
    }

    return Right(JSONRPC2_0ResponseObject(
      result: optionOf<dynamic>(map[JSON_RPC_Constants.RESULT_KEY]),
      error: optionOf<dynamic>(map[JSON_RPC_Constants.ERROR_KEY])
          .map((dynamic a) => JSON_RPC_2_Error.fromJSON(a)),
      id: map[JSON_RPC_Constants.ID_KEY],
    ));
  }

  static Either<RPCException, JSONRPC2_0ResponseObject> fromString(String str) {
    try {
      return fromJson(json.decode(str) as Map<String, dynamic>);
    } catch (e) {
      return left(BadDataRPCException(e.toString()));
    }
  }

  Map<String, dynamic> toJson() {
    final Map<String, dynamic> json = <String, dynamic>{
      JSON_RPC_Constants.JSONRPC_KEY:
          JSON_RPC_Constants.JSON_RPC_2_0_VERSION_VALUE,
      JSON_RPC_Constants.ID_KEY: id,
    };

    result.fold(() {}, (dynamic result) {
      json[JSON_RPC_Constants.RESULT_KEY] = result;
    });

    error.fold(() {}, (error) {
      json[JSON_RPC_Constants.ERROR_KEY] = error.toJson();
    });

    return json;
  }

  @override
  String toString() => json.encode(toJson());
}

/// When there's no result or error inside an rpc response object.
class NoContentRPCException implements RPCException {
  final String cause;

  const NoContentRPCException(this.cause);

  @override
  String toString() => '$NoContentRPCException{cause: $cause}';
}

/// When the response object doesn't have an id.
class NoIDRPCException implements RPCException {
  final String cause;

  const NoIDRPCException(this.cause);

  @override
  String toString() => '$NoIDRPCException{cause: $cause}';
}

/// The error codes from and including -32768 to -32000 are reserved for pre-defined errors.
class JSON_RPC_2_Error implements RPCException {
  static const String CODE_KEY = "code";
  static const String MESSAGE_KEY = "message";
  static const String DATA_KEY = "data";

  final int code;
  final String message;
  final Option<Object> data;

  const JSON_RPC_2_Error(this.code, this.message, this.data);

  /// Invalid JSON was received by the server.
  /// An error occurred on the server while parsing the JSON text.
  factory JSON_RPC_2_Error.parseError({Option<Object> data = const None()}) =>
      JSON_RPC_2_Error(-32700, "Parse error", data);

  /// The JSON sent is not a valid Request object.
  factory JSON_RPC_2_Error.invalidRequest(
          {Option<Object> data = const None()}) =>
      JSON_RPC_2_Error(-32600, "Invalid Request", data);

  /// The method does not exist / is not available.
  factory JSON_RPC_2_Error.methodNotFound(
          {Option<Object> data = const None()}) =>
      JSON_RPC_2_Error(-32601, "Method not found", data);

  /// Invalid method parameter(s).
  factory JSON_RPC_2_Error.invalidParams(
          {Option<Object> data = const None()}) =>
      JSON_RPC_2_Error(-32602, "Invalid params", data);

  /// Internal JSON-RPC error.
  factory JSON_RPC_2_Error.internalError(
          {Option<Object> data = const None()}) =>
      JSON_RPC_2_Error(-32603, "Internal error", data);

  /// Reserved for implementation-defined server-errors.
  ///
  /// [errorCodeOffset] can lie between 0 and 99.
  /// A value of 0 will result in the error code -32000
  /// A value of 1 will result in the error code -32001
  /// and so on.
  /// as defined all codes from -32000 to -32099 are reserved for this error.
  factory JSON_RPC_2_Error.serverError(int errorCodeOffset,
      {Option<Object> data = const None()}) {
    assert(errorCodeOffset >= 0 && errorCodeOffset <= 99);
    return JSON_RPC_2_Error(-32000 - errorCodeOffset, "Server error", data);
  }

  /// Reserved for implementation-defined server-errors.
  factory JSON_RPC_2_Error.application(int code, String message,
      {Option<Object> data = const None()}) {
    return JSON_RPC_2_Error(code, message, data);
  }

  /// Reserved for implementation-defined server-errors.
  factory JSON_RPC_2_Error.fromJSON(dynamic json) {
    if (json is Map) {
      if (!json.containsKey(CODE_KEY)) {
        return JSON_RPC_2_Error.parseError(
            data: some({"reason": "json does not contain a code"}));
      }

      if (!json.containsKey(MESSAGE_KEY)) {
        return JSON_RPC_2_Error.parseError(
            data: some({"reason": "json does not contain a message"}));
      }

      return JSON_RPC_2_Error(json[CODE_KEY] as int,
          json[MESSAGE_KEY] as String, optionOf(json[DATA_KEY]));
    } else {
      return JSON_RPC_2_Error.parseError(
          data: some({"reason": "json is not a map but ${json.runtimeType}"}));
    }
  }

  dynamic toJson() {
    final kjson = {
      CODE_KEY: code,
      MESSAGE_KEY: message,
    };

    data.fold(() {}, (data) {
      kjson[DATA_KEY] = data;
    });

    return kjson;
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is JSON_RPC_2_Error &&
          runtimeType == other.runtimeType &&
          code == other.code;

  @override
  int get hashCode => code.hashCode;

  @override
  String toString() => json.encode(toJson());
}

/// Some servers may send a mix of request and response RPC objects.
///
/// Use this to convert an incoming RPC message.
///
/// There are 4 possible cases that may occur:
///
/// - [request] the data is a [JSONRPC2_0RequestObject]. Its values are passed to the callback.
///
/// - [response] the data is a [JSONRPC2_0ResponseObject]. Its result is passed to the callback.
///
/// - [error] the data is a [JSONRPC2_0ResponseObject] with an error. Its parsed error object is passed to the callback.
///
/// - [failed] the data could not be parsed. The error object and its stacktrace are passed to the callback.
///   [failed] will be called when:
///   a: the data could not be parsed into a json object.
///   b: the data was neither a [JSONRPC2_0RequestObject] nor a [JSONRPC2_0ResponseObject] object.
///      Footnote: Some servers extend the JSONRPC 2.0 protocol to fit their needs so (b) may also occur when the
///      implemented JSONRPC model representation isn't the one the server expects. Use the provided optional debug
///      parameters to find out if that was the case.
///
/// - [data] A json encoded string of the RPC object.
T mapAnyJsonRpc20Object<T>({
  @required String data,
  @required
      T Function(String method, Option<Object> params, Option<Object> id)
          request,
  @required T Function(Object result, Object id) response,
  @required T Function(JSON_RPC_2_Error error, Object id) error,
  @required T Function(Object error, StackTrace stackTrace) failed,
}) {
  try {
    /// Test: B
    return JSONRPC2_0RequestObject.fromString(data).fold(
        (err) => throw err, (obj) => request(obj.method, obj.params, obj.id));
  } catch (e, _) {
    try {
      Option<Object> res;
      Option<JSON_RPC_2_Error> err;
      dynamic id;

      try {
        final obj = JSONRPC2_0ResponseObject.fromString(data)
            .fold((a) => throw a, (a) => a);
        res = obj.result;
        err = obj.error;
        id = obj.id;
      } on NoContentRPCException catch (_) {
        // This is caught here to make C (below) accessible and therefore
        // covered in a unit test.
        res = const None();
        err = const None();
      } catch (e) {
        rethrow;
      }

      return res.fold(() {
        return err.fold(() {
          /// Test: C
          return failed(
            const NoContentRPCException(
                "neither result nor error were not present"),
            StackTrace.fromString(""),
          );
        }, (err) {
          /// Test: D
          return error(err, id);
        });
      }, (dynamic a) {
        /// Test: E
        return response(a, id);
      });
    } catch (e, s) {
      /// Test: F
      // Documentation@b: The json data was neither a request not a response object.
      return failed(e, s);
    }
  }
}

bool whitelistRpcObject({
  @required String data,
  bool Function(String method, Option<Object> params, Option<Object> id)
      request,
  bool Function(Object result, Object id) response,
  bool Function(JSON_RPC_2_Error error, Object id) error,
  bool Function(Object error, StackTrace stackTrace) failed,
}) =>
    mapAnyJsonRpc20Object(
      data: data,
      // Test: A
      request: request ?? (_, __, ___) => false,
      // Test: B
      response: response ?? (_, __) => false,
      // Test: C
      error: error ?? (_, __) => false,
      // Test: D
      failed: failed ?? (_, __) => false,
    );

bool whitelistRequestRpcObject({
  @required String data,
  String method,
  bool Function(Object) allowParams,
}) =>
    mapAnyJsonRpc20Object(
      data: data,
      // Test: A
      request: (_method, params, id) =>
          _method == method && allowParams(params.getOrElse(() => null)),
      // Test: B
      response: (_, __) => false,
      // Test: C
      error: (_, __) => false,
      // Test: D
      failed: (_, __) => false,
    );
