import 'dart:convert';

import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:meta/meta.dart';

/// Cache for json data converted from string into json.
///
/// Used for optimization because many listeners may need json data
/// but it should not be converted multiple times.
///
/// Call [str] to get the raw string or [cachedJson] to get the json representation.
class StringConversion {
  final String str;

  dynamic _json;

  StringConversion(this.str);

  /// Converts and stores the json decoded [str] value.
  ///
  /// Converts it only once if it hasn't already and subsequent calls return the first conversion.
  dynamic cachedJson() {
    if (_json != null) {
      return _json;
    } else {
      return _json = json.decode(str);
    }
  }

  // ignore: avoid_returning_this
  StringConversion apply<T>(void Function(StringConversion) execute) {
    execute(this);
    return this;
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is StringConversion &&
              runtimeType == other.runtimeType &&
              str == other.str;

  @override
  int get hashCode => str.hashCode;
}

/// Base parsing error
///
/// Used within a [OptionalParsedResult]
abstract class ParsingError {
  const ParsingError._();

  R map<R>({
    @required R Function(ThrowsWhileParsingError) throwsWhileParsing,
    @required R Function(ParsingValidationError) parsingValidation,
  }) {
    final self = this;
    if (self is ThrowsWhileParsingError) {
      return throwsWhileParsing(self);
    }
    if (self is ParsingValidationError) {
      return parsingValidation(self);
    }
    throw Exception("unhandled");
  }
}

APIErrors parsingErrorToUnknownAPIError(ParsingError err) {
  return err.map(
      parsingValidation: (parsing) {
        return UnknownAPIError(Some(parsing.check.description), none());
      }, throwsWhileParsing: (validation) {
    return UnknownAPIError(
        Some(validation.describe()), Some(validation.stackTrace));
  });
}

/// Used when an error was caught during parsing.
///
/// [stackTrace] is not counted to its identity.
class ThrowsWhileParsingError extends ParsingError {
  final dynamic error;

  final StackTrace stackTrace;

  final ThrowsAtStage stage;

  final String stageDescription;

  final String source;

  const ThrowsWhileParsingError(this.error, this.stackTrace, this.stage,
      this.stageDescription, this.source) : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is ThrowsWhileParsingError &&
              runtimeType == other.runtimeType &&
              error == other.error &&
              stage == other.stage;

  @override
  int get hashCode => error.hashCode ^ stage.hashCode;

  @override
  String toString() =>
      '$ThrowsWhileParsingError{\n\terror: $error, \n\tstackTrace: $stackTrace, \n\tstage: $stage, \n\t: stageDescription: $stageDescription, \n\tsource: $source\n}';

  String describe() => toString();
}

enum ThrowsAtStage {
  preInterestingCheck,
  parsing,
  postInterestingCheck,
  validation,
}

/// Used when a [ValidationCheck] has failed while parsing.
///
/// Does not count [check] to its identity.
class ParsingValidationError<T> extends ParsingError {
  final ValidationCheck<T> check;

  final T value;

  const ParsingValidationError(this.check, this.value) : super._();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is ParsingValidationError &&
              runtimeType == other.runtimeType &&
              value == other.value;

  @override
  int get hashCode => value.hashCode;

  @override
  String toString() =>
      '$ParsingValidationError{\n\tcheck: ${check
          ?.description}, \n\tvalue: $value\n}';
}

/// If [isValid] returns false then when used in an [OptionalParsedResult]
/// the result will give be a [ParsingValidationError].
class ValidationCheck<T> {
  final bool Function(T) isValid;

  final String description;

  const ValidationCheck(this.isValid, this.description);

  @override
  String toString() =>
      '$ValidationCheck($runtimeType $T){\n\tdescription: $description\n}';
}

/// Callbacks with a description to tell whether a [StringConversion] is of interest or not.
class PreInterestCheck {
  final bool Function(StringConversion) isInteresting;

  final String description;

  const PreInterestCheck(this.isInteresting, this.description);

  @override
  String toString() =>
      '$PreInterestCheck($runtimeType){\n\tdescription: $description\n}';
}

/// Callbacks with a description to tell whether Type [T] is of interest or not.
class PostInterestCheck<T> {
  final bool Function(T) isInteresting;

  final String description;

  const PostInterestCheck(this.isInteresting, this.description);

  @override
  String toString() =>
      '$PostInterestCheck($runtimeType $T){\n\tdescription: $description\n}';
}

/// Use this to parse values of type [T].
///
/// Additionally
/// - You may specify conditions before the value is parsed to not even start parsing.
/// - You may specify conditions after the value is parsed to not even start validation.
/// - you may specify whether the parsed value should be accepted by using validators and in case it fails an error will be returned.
/// - All exceptions are caught for each stage separately to make debugging easier.
class OptionalParsedResult<T> {
  final Option<Either<ParsingError, T>> _value;

  Option<Either<ParsingError, T>> get value => _value;

  static OptionalParsedResult<T> Function(String) str<T>(
      T Function(StringConversion conversion) converter, {
        List<PreInterestCheck> preInteresting = const [],
        List<PostInterestCheck<T>> postInteresting = const [],
        List<ValidationCheck<T>> validations = const [],
      }) {
    return (str) {
      return OptionalParsedResult<T>(
        source: StringConversion(str),
        converter: converter,
        preInteresting: preInteresting,
        postInteresting: postInteresting,
        validations: validations,
      );
    };
  }

  OptionalParsedResult({
    /// Source data
    @required StringConversion source,

    /// Raw conversion from [conversion] to [T].
    ///
    /// Errors thrown here will be caught but should not be relied to as a way of validating
    /// the model. Use other filtering stages for that.
    @required T Function(StringConversion conversion) converter,

    /// Filter the raw [StringConversion] for channelName or other characteristics.
    /// Because all messages are generally put through a [OptionalParsedResult] this
    /// should catch any message that is not meant to  be parsed by this [OptionalParsedResult].
    ///
    /// Prefer to do it here and not in [postInteresting] or [validators] because conversions will have
    /// taken place there already and for potentially large jsons this could mean performance problems.
    ///
    /// If you fail to filter messages so that they are only received by their intended recipients
    /// weird behavior might happen. For example when 2 models overlap with what they accept, this could
    /// mean that a conversion could be seen as being 'correctly converter' even though some
    /// content was dropped or never initialized. This could be caught by [validators] but should
    /// generally not be needed if the correct filtered takes place here.
    List<PreInterestCheck> preInteresting = const [],

    /// Filter out results by their parsed state.
    /// This should signal expected behavior.
    List<PostInterestCheck<T>> postInteresting = const [],

    /// Validate that the parsed result contains what it needs to contain to be useful.
    /// This should act as a safeguard and signal unexpected behavior.
    List<ValidationCheck<T>> validations = const [],
  }) : this._value = tryPreInteresting<T>(source, preInteresting) ??
      tryConversion<T>(() => converter(source), source.str).fold(
              (a) =>
          tryPostInteresting<T>(a, postInteresting) ??
              tryValidation<T>(a, validations),
          id);

  const OptionalParsedResult.none() : _value = const None();

  @visibleForTesting
  const OptionalParsedResult.raw(this._value);

  @visibleForTesting
  static Option<Either<ParsingError, T>> tryPreInteresting<T>(
      StringConversion source, List<PreInterestCheck> preInteresting) {
    bool isPreInteresting = true;
    final preInterestingIterator = preInteresting.iterator;

    try {
      while (preInterestingIterator.moveNext() && isPreInteresting) {
        final check = preInterestingIterator.current.isInteresting(source);
        if (!check) {
          isPreInteresting = false;
        }
      }
    } catch (e, s) {
      return Some(Left(ThrowsWhileParsingError(
          e,
          s,
          ThrowsAtStage.preInterestingCheck,
          preInterestingIterator.current.toString(),
          source.str)));
    }

    if (!isPreInteresting) {
      return none();
    } else {
      return null;
    }
  }

  @visibleForTesting
  static Either<T, Option<Either<ParsingError, T>>> tryConversion<T>(
      T Function() convert, String debugSource) {
    T a;

    try {
      a = convert();
    } catch (e, s) {
      return Right(Some(Left(ThrowsWhileParsingError(e, s,
          ThrowsAtStage.parsing, "Failed parsing Type $T", debugSource))));
    }

    return Left(a);
  }

  @visibleForTesting
  static Option<Either<ParsingError, T>> tryPostInteresting<T>(T value,
      List<PostInterestCheck<T>> postInteresting) {
    bool isPostInteresting = true;
    final postInterestingIterator = postInteresting.iterator;
    try {
      while (postInterestingIterator.moveNext() && isPostInteresting) {
        final check = postInterestingIterator.current.isInteresting(value);
        if (!check) {
          isPostInteresting = false;
        }
      }
    } catch (e, s) {
      return Some<Either<ParsingError, T>>(Left<ParsingError, T>(
          ThrowsWhileParsingError(e, s, ThrowsAtStage.postInterestingCheck,
              postInterestingIterator.current.toString(), value.toString())));
    }

    if (!isPostInteresting) {
      return none();
    } else {
      return null;
    }
  }

  @visibleForTesting
  static Option<Either<ParsingError, T>> tryValidation<T>(T value,
      List<ValidationCheck<T>> validations) {
    ValidationCheck<T> failedCheck;
    final validationsIterator = validations.iterator;

    try {
      while (validationsIterator.moveNext() && failedCheck == null) {
        final check = validationsIterator.current.isValid(value);
        if (!check) {
          failedCheck = validationsIterator.current;
        }
      }
    } catch (e, s) {
      return Some(
          Left(ThrowsWhileParsingError(e, s, ThrowsAtStage.validation,
              validationsIterator.toString(), value.toString())));
    }

    if (failedCheck == null) {
      return Some(Right(value));
    } else {
      return Some(Left(ParsingValidationError<T>(failedCheck, value)));
    }
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is OptionalParsedResult<T> &&
              runtimeType == other.runtimeType &&
              _value == other._value;

  @override
  int get hashCode => _value.hashCode;

  Option<T> get toOption => _value.bind((a) => a.toOption());

  @override
  String toString() => '${OptionalParsedResult}($T){\n\tvalue: $value\n}';

  static Option<Either<AllErrors, T>> lift<T>(
      Option<Either<AllErrors, OptionalParsedResult<T>>> result) {
    return optionOf(result).bind(id).bind((result) {
      return result.fold(
        someLeft,
            (result) {
          return result.value.bind((either) {
            return either.fold((parsingError) {
              return someLeft(parsingErrorToUnknownAPIError(parsingError));
            }, someRight);
          });
        },
      );
    });
  }
}

/// TODO create and move to an EitherOption
Option<Either<L, R>> someLeft<L, R>(L l) => some(left(l));

/// TODO create and move to an EitherOption in dartz
Option<Either<L, R>> someRight<L, R>(R l) => some(right(l));
