import 'dart:async';

import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:mockito/mockito.dart';
import 'package:statemachine/statemachine.dart';
import 'package:test/test.dart';

void main() {
    group("$ReducerStateInterface", () {
        /// Nothing to test
    });

    group("$SocketChannelReducer", () {
        /// Nothing to test
    });

    test("$SocketReducerAction", () {
        expect(() => TestSocketReducerAction().map<dynamic>(identity: null, connected: null),
            throwsA(const TypeMatcher<Exception>()));
    });

    test("$SocketReducerActionIdentity", () {
        const SocketReducerActionIdentity a = SocketReducerActionIdentity();
        expect(a.map(identity: () => true, connected: () => false), true);
    });

    test("$SocketReducerActionConnected", () {
        const SocketReducerActionConnected a = SocketReducerActionConnected();
        expect(a.map(identity: () => false, connected: () => true), true);
    });

    group("$DebugSocketChannelReducer", () {
        test("register prinnts", () {
            bool printed = false;

            runZoned(() {
                DebugSocketChannelReducer("-").register(null);
            }, zoneSpecification: ZoneSpecification(print: (self, parent, zone, str) {
                printed = true;
            }));

            expect(printed, isTrue);
        });

        test("messageReducer prints", () {
            bool printed = false;

            runZoned(() {
                DebugSocketChannelReducer("-").messageReducer(
                    const APIStateConnected(),
                    StringConversion("-"),
                );
            }, zoneSpecification: ZoneSpecification(print: (self, parent, zone, str) {
                printed = true;
            }));

            expect(printed, isTrue);
        });

        test("dispose prints", () {
            bool printed = false;

            runZoned(() {
                DebugSocketChannelReducer("-").dispose();
            }, zoneSpecification: ZoneSpecification(print: (self, parent, zone, str) {
                printed = true;
            }));

            expect(printed, isTrue);
        });
    });

    group("$ConnectedChannelReducer", () {
        group("messageReducer", () {
            test("dont return connected on false message", () {
                expect(
                    ConnectedChannelReducer((str) => str.str == "1")
                        .messageReducer(const APIStateLoading(), StringConversion("0")),
                    const SocketReducerActionIdentity(),
                );
            });

            test("dont return connected on false state", () {
                expect(
                    ConnectedChannelReducer((str) => str.str == "1")
                        .messageReducer(const APIStateInitialized(), StringConversion("1")),
                    const SocketReducerActionIdentity(),
                );
            });

            test("return connected on correct state and message", () {
                expect(
                    ConnectedChannelReducer((str) => str.str == "1")
                        .messageReducer(const APIStateLoading(), StringConversion("1")),
                    const SocketReducerActionConnected(),
                );
            });
        });

        test("dispose does nothing", () {
            ConnectedChannelReducer((_) => true).dispose();
        });
    });

//    group("$TimeoutChannelReducer", () {
//        group("messageReducer", () {
//            test("emit disconnected after time runs out", () async {
//                BehaviorSubject<APIState> state = BehaviorSubject.seeded(const APIStateInitialized());
//
//                expect(state, emitsInOrder(<dynamic>[
//                    const APIStateInitialized(),
//                    const TypeMatcher<APIStateDisconnected>(),
//                    emitsDone,
//                ]));
//
//                final reducer = TimeoutChannelReducer(period: const Duration(milliseconds: 50));
//
//                reducer.register(_TestReducerStateInterface(
//                        (_) => throw Exception("should not be called"),
//                        (e) => state.add(APIStateDisconnected(Some(e)))
//                ));
//
//                reducer.messageReducer(const APIStateConnected(), StringConversion("a"));
//                await Future<void>.delayed(const Duration(milliseconds: 30));
//
//                reducer.messageReducer(const APIStateConnected(), StringConversion("a"));
//                await Future<void>.delayed(const Duration(milliseconds: 30));
//
//                reducer.messageReducer(const APIStateConnected(), StringConversion("a"));
//                await Future<void>.delayed(const Duration(milliseconds: 60));
//
//                await reducer.dispose();
//                await state.close();
//            });
//
//            test("don't emit disconnected before timer runs out", () async {
//                BehaviorSubject<APIState> state = BehaviorSubject.seeded(const APIStateInitialized());
//
//                expect(state, emitsInOrder(<dynamic>[
//                    const APIStateInitialized(),
//                    emitsDone,
//                ]));
//
//                final reducer = TimeoutChannelReducer(period: const Duration(milliseconds: 50));
//
//                reducer.register(_TestReducerStateInterface(
//                        (_) => throw Exception("should not be called"),
//                        (a) => state.add(APIStateDisconnected(some(a))),
//                ));
//
//                expect(reducer.messageReducer(const APIStateConnected(), StringConversion("a")), const SocketReducerActionIdentity());
//
//                await Future<void>.delayed(const Duration(milliseconds: 30));
//
//                await reducer.dispose();
//                await state.close();
//            });
//
//            test("forgot to call register throws", () async {
//                Object obj;
//
//                runZoned(
//                        () {
//                        final reducer = TimeoutChannelReducer(period: const Duration(milliseconds: 20));
//                        reducer.messageReducer(const APIStateConnected(), StringConversion("a"));
//                    },
//                    zoneSpecification: ZoneSpecification(
//                        handleUncaughtError: (a, b, c, object, stacktrace) => obj = object,
//                    ),
//                );
//
//                await Future<void>.delayed(const Duration(milliseconds: 30));
//
//                expect(obj, const TypeMatcher<Exception>());
//            });
//        });
//
//        group("dispose", () {
//            test("cancels the timer", () async {
//                final reducer = TimeoutChannelReducer(period: const Duration(milliseconds: 1000));
//                reducer.t = Timer(reducer.period, () {});
//
//                await reducer.dispose();
//
//                expect(reducer.t.isActive, isFalse);
//            });
//        });
//    });

    group("$MarcopoloChannelReducer", () {
        group("messageReducer", () {
            test("throws when not registered", () {
                MarcopoloChannelReducer reducer;

                reducer = MarcopoloChannelReducer<dynamic>((str) {
                    if (str.str == "1") {
                      return some("one");
                    }
                    if (str.str == "2") {
                      return some("two");
                    } else {
                      return none();
                    }
                });

                expect(() {
                    return reducer.messageReducer(const APIStateConnected(), StringConversion("message"));
                }, throwsA(const TypeMatcher<Exception>()));
            });

            test("sends messages", () {
                MarcopoloChannelReducer reducer;

                reducer = MarcopoloChannelReducer<dynamic>((str) {
                    if (str.str == "1") {
                      return some("one");
                    }
                    if (str.str == "2") {
                      return some("two");
                    } else {
                      return none();
                    }
                });
                final Signal<String> a = Signal("");

                expect(a.wave.skip(1).asStream(), emitsInOrder(<dynamic>[
                    "one",
                    "two",
                    emitsDone,
                ]));

                reducer.register(_TestReducerStateInterface(
                    a.add, (_) => throw Exception("should not be called"),
                ));
                reducer.messageReducer(const APIStateInitialized(), StringConversion("1"));
                reducer.messageReducer(const APIStateConnected(), StringConversion("1"));
                reducer.messageReducer(const APIStateConnected(), StringConversion("2"));
                reducer.messageReducer(const APIStateConnected(), StringConversion("0"));

                a.close();
            });
        });

        test("dispose", () {
            /// Does nothing
            MarcopoloChannelReducer<dynamic>((a) => none()).dispose();
        });
    });

    group("$ModelCallbackChannelReducer", () {
        test("messageReducer", () {
            int t;

            ModelCallbackChannelReducer<int>(
                converter: (a) => a.str.length,
                callback: (int a) => t = a,
            ).messageReducer(const APIStateInitialized(), StringConversion("aaaa"));

            expect(t, 4);
        });

        test("dispose", () {
            /// dispose does nothing.
            ModelCallbackChannelReducer<int>(
                converter: (a) => a.str.length,
                callback: (int a) {},
            ).dispose();
        });
    });

    group("$runReducers", () {
        test("identity", () {
            final List<APIState> states = [];
            final _MockSocketChannelReducer reducer = _MockSocketChannelReducer();

            when(reducer.messageReducer(const APIStateLoading(), StringConversion("abc"))).thenReturn(
                const SocketReducerActionIdentity());

            runReducers([
                reducer,
            ], StringConversion("abc"), const APIStateLoading(), states.add);

            verify(reducer.messageReducer(const APIStateLoading(), StringConversion("abc")));

            expect(states, isEmpty);
        });

        test("connected", () {
            final List<APIState> states = [];
            final _MockSocketChannelReducer reducer = _MockSocketChannelReducer();

            when(reducer.messageReducer(const APIStateLoading(), StringConversion("abc"))).thenReturn(
                const SocketReducerActionConnected());

            runReducers([
                reducer,
            ], StringConversion("abc"), const APIStateLoading(), states.add);

            verify(reducer.messageReducer(const APIStateLoading(), StringConversion("abc")));

            expect(states, [
                const APIStateConnected(),
            ]);
        });
    });
}

class _MockSocketChannelReducer extends Mock implements SocketChannelReducer {}

class CommonConnectedStateTestHelper<T> {
    AllErrors hasDisconnected;
    int builtSockets;
    MockWebSocket socket;
    MockSocketChannelReducer reducer;
    Machine machine;

    MockSocketChannelPreferences prefs;

    Signal<Option<StringConversion>> receivedMessages; // ignore: close_sinks
    Signal<Option<Either<AllErrors, SocketInterface>>> lastData; // ignore: close_sinks
    Signal<Option<Either<AllErrors, T>>> lastTData; // ignore: close_sinks
    Signal<APIState> lastState; // ignore: close_sinks

    void init() {
        builtSockets = 0;
        hasDisconnected = null;
        lastData = Signal(const None());
        lastTData = Signal(const None());
        lastState = Signal(const APIStateInitialized());
        reducer = MockSocketChannelReducer();
        receivedMessages = Signal(const None());
        when(reducer.dispose()).thenAnswer((_) async {});
        machine = Machine();
        prefs = MockSocketChannelPreferences();
        socket = MockWebSocket();
    }
}

class MockSocketChannelReducer extends Mock implements SocketChannelReducer {}

class MockWebSocket extends SocketInterface {

    final Signal msgs = Signal<dynamic>(null);

    @override
    Future close() async {
        kstream.close();
        msgs.close();
    }

    @override
    void add(String str) => msgs.add(str);

    Signal<Option<Either<AllErrors, String>>> kstream = Signal(const None<Either<AllErrors, String>>());

    @override
    // ignore: close_sinks
    Wave<Option<Either<AllErrors, String>>> get stream => kstream.wave;
}

class MockSocketChannelPreferences extends Mock implements SocketChannelPreferences {}

class TestSocketReducerAction extends SocketReducerAction {}

class _TestReducerStateInterface extends ReducerStateInterface {

    void Function(String) sMsg;

    void Function(AllErrors) disco;

    _TestReducerStateInterface(this.sMsg, this.disco);

    @override
    void sendMessage(String msg) => sMsg(msg);

    @override
    void disconnect(AllErrors error) => disco(error);
}