import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import '../service_endpoint_spec.dart';
import 'reducer_test.dart';

void main() {
    group("SocketChannelServiceEndpoint", () {
        group("SPEC", () {
            MockSocketChannelPreferences prefs() {
                SocketChannelReducer reducer;
                MockSocketChannelPreferences preferences;

                reducer = MockSocketChannelReducer();
                preferences = MockSocketChannelPreferences();

                when(reducer.messageReducer(any, any)).thenAnswer((a) => const SocketReducerActionIdentity());
                when(reducer.dispose()).thenAnswer((a) async {});

                when(preferences.onReceivedSocket).thenReturn(none());
                when(preferences.onLostSocket).thenReturn(none());
                when(preferences.reducers).thenReturn([reducer]);

                return preferences;
            }

            MockSocketEndpoint socket([APIStatusError error]) {
                final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                final MockWebSocket webSocket = MockWebSocket();
                // ignore: close_sinks
                final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                = Signal(Some(error != null ? left(error) : right(webSocket)));
                // ignore: close_sinks
                final Signal<Option<StringConversion>> socketData
                = Signal(const None());

                when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                return socketEndpoint;
            }

            test("${ServiceEndpointVerifier.stateConstruct}", () async {
                await ServiceEndpointVerifier.stateConstruct(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateConstructAndDispose}", () async {
                await ServiceEndpointVerifier.stateConstructAndDispose(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenOnce}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenReceivesError}", () async {
                await ServiceEndpointVerifier.stateListenReceivesError(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(statusError(1234)),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.stateUnlistenInitializes}", () async {
                await ServiceEndpointVerifier.stateUnlistenInitializes(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenTwice}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeWhileConnected}", () async {
                await ServiceEndpointVerifier.stateDisposeWhileConnected(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting}", () async {
                await ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeAfterUnlistening}", () async {
                await ServiceEndpointVerifier.stateDisposeAfterUnlistening(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners}", () async {
                final List<MockSocketEndpoint> createdSockets = [
                        () {
                        final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                        final MockWebSocket webSocket = MockWebSocket();
                        // ignore: close_sinks
                        final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                        = Signal(Some(right(webSocket)));
                        // ignore: close_sinks
                        final Signal<Option<StringConversion>> socketData
                        = Signal(Some(StringConversion("111")));

                        when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                        when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                        return socketEndpoint;
                    }(),
                        () {
                        final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                        // ignore: close_sinks
                        final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                        = Signal(const Some(Left(noInternetError)));
                        // ignore: close_sinks
                        final Signal<Option<StringConversion>> socketData
                        = Signal(Some(StringConversion("111")));

                        when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                        when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                        return socketEndpoint;
                    }(),
                ];

                await ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners(SocketChannelServiceEndpoint<int>(
                    ownsSocket: false,
                    socket: (_) => createdSockets.removeLast(),
                    channel: prefs(),
                    converter: (str) => str.str.length,
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                ));
            });

            test("${ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners}", () async {
                final List<MockSocketEndpoint> createdSockets = [
                        () {
                        final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                        final MockWebSocket webSocket = MockWebSocket();
                        // ignore: close_sinks
                        final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                        = Signal(Some(right(webSocket)));
                        // ignore: close_sinks
                        final Signal<Option<StringConversion>> socketData
                        = Signal(Some(StringConversion("111")));

                        when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                        when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                        return socketEndpoint;
                    }(),
                        () {
                        final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                        // ignore: close_sinks
                        final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                        = Signal(const Some(Left(noInternetError)));
                        // ignore: close_sinks
                        final Signal<Option<StringConversion>> socketData
                        = Signal(Some(StringConversion("111")));

                        when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                        when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                        return socketEndpoint;
                    }(),
                ];

                await ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners(SocketChannelServiceEndpoint<int>(
                    ownsSocket: false,
                    socket: (_) => createdSockets.removeLast(),
                    channel: prefs(),
                    converter: (str) => str.str.length,
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                ));
            });

            test("${ServiceEndpointVerifier.stateReconnectWithNoListeners}", () async {
                await ServiceEndpointVerifier.stateReconnectWithNoListeners(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataConstruct}", () async {
                await ServiceEndpointVerifier.dataConstruct(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataSingleListen}", () async {
                await ServiceEndpointVerifier.dataSingleListen(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) =>
                                () {
                                final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                                final MockWebSocket webSocket = MockWebSocket();
                                // ignore: close_sinks
                                final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                                = Signal(Some(right(webSocket)));
                                // ignore: close_sinks
                                final Signal<Option<StringConversion>> socketData
                                = Signal(Some(StringConversion("111")));

                                when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                                when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                                return socketEndpoint;
                            }(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    3,
                );
            });

            test("${ServiceEndpointVerifier.dataSingleListenError}", () async {
                await ServiceEndpointVerifier.dataSingleListenError(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(statusError(1234)),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.dataConstructDispose}", () async {
                await ServiceEndpointVerifier.dataConstructDispose(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataListenSuccessDispose}", () async {
                await ServiceEndpointVerifier.dataListenSuccessDispose(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) =>
                                () {
                                final MockSocketEndpoint socketEndpoint = MockSocketEndpoint();
                                final MockWebSocket webSocket = MockWebSocket();
                                // ignore: close_sinks
                                final Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData
                                = Signal(Some(right(webSocket)));
                                // ignore: close_sinks
                                final Signal<Option<StringConversion>> socketData
                                = Signal(Some(StringConversion("four")));

                                when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                                when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);

                                return socketEndpoint;
                            }(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    4,
                );
            });

            test("${ServiceEndpointVerifier.dataListenErrorDispose}", () async {
                await ServiceEndpointVerifier.dataListenErrorDispose(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(statusError(1234)),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.dataDispose}", () async {
                await ServiceEndpointVerifier.dataDispose(
                    SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socket(),
                        channel: prefs(),
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });
        });

        group("", () {
            MockSocketEndpoint socketEndpoint;
            MockSocketChannelPreferences preferences;
            SocketChannelReducer reducer;
            Signal<Option<Either<AllErrors, SocketInterface>>> socketLastData; // ignore: close_sinks
            MockWebSocket webSocket;

            Signal<Option<StringConversion>> socketData; // ignore: close_sinks

            SocketChannelServiceEndpoint<int> sut;

            setUp(() {
                socketEndpoint = MockSocketEndpoint();
                reducer = MockSocketChannelReducer();
                webSocket = MockWebSocket();
                preferences = MockSocketChannelPreferences();
                socketData = Signal(const None());
                socketLastData = Signal(Some(right(webSocket)));

                when(reducer.messageReducer(any, any)).thenAnswer((a) => const SocketReducerActionIdentity());
                when(reducer.dispose()).thenAnswer((a) async {});
                when(socketEndpoint.lastData).thenAnswer((_) => socketLastData.wave);
                when(socketEndpoint.socketData).thenAnswer((_) => socketData.wave);
                when(preferences.onReceivedSocket).thenReturn(none());
                when(preferences.onLostSocket).thenReturn(none());
                when(preferences.reducers).thenReturn([reducer]);

                sut = SocketChannelServiceEndpoint<int>(
                    ownsSocket: false,
                    socket: (_) => socketEndpoint,
                    channel: preferences,
                    converter: (str) => str.str.length,
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                );
            });

            test("initializes all states and starts out with the initialization state", () {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    const APIStateInitialized(),
                ]));
                expect(sut.lastState.value, apiInitialized);
            });

            test("diconnect", () async {
                sut.disconnect(APIStatusError(some(123), none()));
                await pumpEventQueue();
                expect(sut.lastState.value, APIStateDisconnected(Some(APIStatusError(some(123), none()))));
                expect(sut.lastData.value, Some(Left<dynamic, dynamic>(APIStatusError(some(123), none()))));
                expect(sut.lastState.value, const TypeMatcher<APIStateDisconnected>());
            });

            test("dont reconnect if no subscribers", () {
                sut.reconnect();
                expect(sut.lastState.value, apiInitialized);
            });

            test("reconnect only if has subscribers", () {
                sut.lastData.subscribe();
                sut.reconnect();
                expect(sut.lastState.value, apiConnected);
            });

            test("dont' treconnect if connected", () {
                sut.lastData.subscribe();
                final instance = sut.lastState.value;
                sut.reconnect();
                expect(identical(sut.lastState.value, instance), true);
            });

            group("verify", () {
                test("firstListen", () async {
                    await ServiceEndpointVerifier.stateListenTwice(sut);
                    await pumpEventQueue();
                    expect(sut.lastState.value, apiConnected);
                });

                test("unlistening changes state to initialized", () async {
                    await ServiceEndpointVerifier.stateUnlistenInitializes(sut);
                    expect(sut.lastState.value, apiInitialized);
                });

                test("manual disconnect > changes state to initialized and distributes $CancelledAPIError error", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        const APIStateInitialized(),
                        const APIStateLoading(),
                        const APIStateConnected(),
                        apiDisconnected(cancelledError),
                    ]));

                    expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<Either>(),
                        const Some(Left<dynamic, dynamic>(CancelledAPIError())),
                    ]));

                    await pumpEventQueue();
                    sut.disconnect(cancelledError);
                    expect(sut.lastState.value, const TypeMatcher<APIStateDisconnected>());
                });

                test("reconnect with no listeners and not in disconnected state does nothing", () async {
                    await ServiceEndpointVerifier.stateReconnectWithNoListeners(sut);
                });
            });
            group("state transitions", () {
                test(
                    "initialized > connected > initialized | once all listeners stop listeners statemachine goes to its initialized state", () async {
                    final SocketChannelServiceEndpoint<int> sut = SocketChannelServiceEndpoint<int>(
                        ownsSocket: false,
                        socket: (_) => socketEndpoint,
                        channel: preferences,
                        converter: (str) => str.str.length,
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    );

                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        const APIStateInitialized(),
                    ]));

                    final sub = sut.lastData.subscribe();

                    await pumpEventQueue();

                    sub.cancel();
                });
            });

            test("reducer registered", () {
                verify(reducer.register(any)).called(1);
            });

            test("connected > dispose", () async {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    ...apiInitLoadConnect,
                    const APIStateDisconnected(Some(CancelledAPIError())),
                ]));

                bool listenerDone = false;

                sut.lastData.subscribe((d) {}, () {
                    listenerDone = true;
                });
                await pumpEventQueue();

                await sut.dispose();

                await pumpEventQueue();

                expect(listenerDone, true);
            });

            test("dispose owned socket", () async {
                final SocketChannelServiceEndpoint<int> sut = SocketChannelServiceEndpoint<int>(
                    ownsSocket: true,
                    socket: (_) => socketEndpoint,
                    channel: preferences,
                    converter: (str) => str.str.length,
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                );

                sut.lastData.subscribe();
                sut.onConnected();
                await sut.dispose();

                await pumpEventQueue();

                verify(socketEndpoint.dispose()).called(1);
            });

            test("sendMessage", () async {
                final SocketChannelServiceEndpoint<int> sut = SocketChannelServiceEndpoint<int>(
                    ownsSocket: true,
                    socket: (_) => socketEndpoint,
                    channel: preferences,
                    converter: (str) => str.str.length,
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                );

                sut.lastData.subscribe();

                await pumpEventQueue();

                sut.sendMessage("abc");

                verify(socketEndpoint.sendMessage("abc")).called(1);
            });
        });
    });
}

class MockSocketEndpoint extends Mock implements SocketEndpointInterface {}
