import 'dart:async';

import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';
import 'package:mockito/mockito.dart';
import 'package:pedantic/pedantic.dart';
import 'package:test/test.dart';

import '../../../../../wave_test.dart';
import '../service_endpoint_spec.dart';
import 'reducer_test.dart';

// ignore_for_file: invalid_use_of_protected_member

void main() {
    group("$SocketEndpointInterface", () {
        /// Nothing to test
    });

    group("SocketEndpointImpl", () {
        group("SPEC", () {
            List<SocketChannelReducer> mockReducer() {
                final SocketChannelReducer reducer = MockSocketChannelReducer();
                when(reducer.messageReducer(any, any)).thenAnswer((a) => const SocketReducerActionIdentity());
                when(reducer.dispose()).thenAnswer((a) async {});
                return [reducer];
            }

            test("${ServiceEndpointVerifier.stateConstruct}", () async {
                await ServiceEndpointVerifier.stateConstruct(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateConstructAndDispose}", () async {
                await ServiceEndpointVerifier.stateConstructAndDispose(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenOnce}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenReceivesError}", () async {
                await ServiceEndpointVerifier.stateListenReceivesError(
                    SocketEndpointImpl(
                        socketCreator: () async {
                            return MockWebSocket()
                                ..kstream.add(Some(Left(statusError(1234))));
                        },
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    statusError(1234),
                );
            });

            test("${ServiceEndpointVerifier.stateUnlistenInitializes}", () async {
                await ServiceEndpointVerifier.stateUnlistenInitializes(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateListenTwice}", () async {
                await ServiceEndpointVerifier.stateListenTwice(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeWhileConnected}", () async {
                await ServiceEndpointVerifier.stateDisposeWhileConnected(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting}", () async {
                await ServiceEndpointVerifier.stateDisposeImmediatelyAfterConnecting(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateDisposeAfterUnlistening}", () async {
                await ServiceEndpointVerifier.stateDisposeAfterUnlistening(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners}", () async {
                final List<MockWebSocket> createdSockets = [
                    MockWebSocket(),
                    MockWebSocket(),
                ];

                createdSockets[0].kstream.add(const Some(Right("data")));
                createdSockets[1].kstream.add(const Some(Left(noInternetError)));

                unawaited(ServiceEndpointVerifier.stateNoInternetOnceThenNoListeners(SocketEndpointImpl(
                    socketCreator: () async => createdSockets.removeLast(),
                    reducers: mockReducer(),
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                )));
            });

            test("${ServiceEndpointVerifier.stateReconnectWithNoListeners}", () async {
                await ServiceEndpointVerifier.stateReconnectWithNoListeners(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataConstruct}", () async {
                await ServiceEndpointVerifier.dataConstruct(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataSingleListen}", () async {
                final data = MockWebSocket();
                await ServiceEndpointVerifier.dataSingleListen(
                    SocketEndpointImpl(
                        socketCreator: () async => data,
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    data,
                );
            });

            /// This doesn't make sense for this implementation.
            /// test("${ServiceEndpointVerifier.dataSingleListenError}", () async {
            ///     await ServiceEndpointVerifier.dataSingleListenError(
            ///         SocketEndpointImpl(
            ///             socketCreator: () async {
            ///                 return MockWebSocket()
            ///                     ..kstream.add(Some(Left(statusError(1234))));
            ///             },
            ///             reducers: mockReducer(),
            ///             whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
            ///         ),
            ///         statusError(1234),
            ///     );
            /// });

            test("${ServiceEndpointVerifier.dataConstructDispose}", () async {
                await ServiceEndpointVerifier.dataConstructDispose(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });

            test("${ServiceEndpointVerifier.dataListenSuccessDispose}", () async {
                final data = MockWebSocket();
                await ServiceEndpointVerifier.dataListenSuccessDispose(
                    SocketEndpointImpl(
                        socketCreator: () async => data,
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                    data,
                );
            });

            /// This doesn't make sense for this implementation.
            /// test("${ServiceEndpointVerifier.dataListenErrorDispose}", () async {
            ///     await ServiceEndpointVerifier.dataListenErrorDispose(
            ///         SocketEndpointImpl(
            ///             socketCreator: () async {
            ///                 return MockWebSocket()
            ///                     ..kstream.add(Some(Left(statusError(1234))));
            ///             },
            ///             reducers: mockReducer(),
            ///             whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
            ///         ),
            ///         null,
            ///     );
            /// });

            test("${ServiceEndpointVerifier.dataDispose}", () async {
                await ServiceEndpointVerifier.dataDispose(
                    SocketEndpointImpl(
                        socketCreator: () async => MockWebSocket(),
                        reducers: mockReducer(),
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    ),
                );
            });
        });

        group("", () {
            SocketChannelReducer reducer;
            SocketEndpointImpl sut;
            List<MockWebSocket> createdSockets;

            setUp(() {
                createdSockets = [];
                reducer = MockSocketChannelReducer();
                when(reducer.messageReducer(any, any)).thenAnswer((a) => const SocketReducerActionIdentity());
                when(reducer.dispose()).thenAnswer((a) async {});
                sut = SocketEndpointImpl(
                    socketCreator: () async {
                        final socket = MockWebSocket();
                        createdSockets.add(socket);
                        return socket;
                    },
                    reducers: [reducer],
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                );
            });

            test("registers reducers", () {
                verify(reducer.register(any)).called(1);
            });

            group("init", () {
                test("inits data to none", () {
                    expect(sut.lastData.value, const None<Either>());
                });
                test("inits state to init", () {
                    expect(sut.lastState.value, const APIStateInitialized());
                });
                test("inits socket data to none", () {
                    expect(sut.socketData.value, const None<Option<StringConversion>>());
                });
                test("inits last state to initialized", () {
                    expect(sut.lastState.value, const APIStateInitialized());
                });
            });

            group("disconnect", () {
                test("state becomes disconnnected if disconnecting directly", () {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        const APIStateInitialized(),
                        apiDisconnected(statusError(1222)),
                    ]));
                    sut.disconnect(statusError(1222));
                });
                test("last data takes on the error if disconnecting directly", () async {
                    expect(sut.lastState.value, const TypeMatcher<APIStateInitialized>());
                    sut.disconnect(APIStatusError(some(1222), none()));
                    expect(sut.lastData.value, Some(Left<AllErrors, SocketInterface>(APIStatusError(some(1222), none()))));
                });
            });

            group("dispose", () {
                test("dispose doesn't change the lastState if lastState is already disconnected", () {
                    expect(sut.lastData.value, const None<Either>());
                    sut.disconnect(statusError(1010));
                    expect(sut.lastData.value, Some<Either<AllErrors, SocketInterface>>(Left(statusError(1010))));
                    expect(sut.lastState.value, apiDisconnected(statusError(1010)));
                    sut.dispose();
                    expect(sut.lastData.value, Some(Left<AllErrors, SocketInterface>(statusError(1010))));
                    expect(sut.lastState.value, apiDisconnected(statusError(1010)));
                });
                test("dispose disposes the reducers", () {
                    sut.dispose();
                    verify(reducer.dispose()).called(1);
                });
            });

            group("connect", () {
                test("subscribing to data changes state to connected", () async {
                    sut.lastData.subscribe();
                    await pumpEventQueue();
                    expect(sut.lastState.value, const APIStateConnected());
                });
                test("subscribing to data initializes a new socket", () {
                    expect(createdSockets.length, 0);
                    sut.lastData.subscribe();
                    expect(createdSockets.length, 1);
                });
                test("subscribing with ConnectImmediatelyPolicy policy connects on the next event queue run", () async {
                    final lastStateTest = WaveTester(sut.lastState);
                    final lastDataTest = WaveTester(sut.lastData);
                    await pumpEventQueue();
                    lastStateTest.expectDidEmitT([
                        const APIStateInitialized(),
                        const APIStateLoading(),
                        const APIStateConnected(),
                    ]);
                    lastDataTest.expectDidEmitT([
                        const None(),
                        Some(Right(createdSockets.first)),
                    ]);
                });
            });

            test("dont't reconnect if connected", () async {
                sut.lastData.subscribe();
                await pumpEventQueue();
                expect(sut.lastState.value, const APIStateConnected());
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    const APIStateConnected(),
                    const APIStateDisconnected(Some(cancelledError)),
                    emitsDone,
                ]));
                sut.reconnect();
                expect(sut.lastState.value, const APIStateConnected());
                await sut.dispose();
            });

            test("reconnect only if disconnected and then refresh socket", () async {
                expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                    ...apiInitLoadConnect,
                    const APIStateDisconnected(Some(cancelledError)),
                    ...apiInitLoadConnect,
                    const APIStateDisconnected(Some(cancelledError)),
                    emitsDone,
                ]));
                final _ = sut.lastData.subscribe();
                await pumpEventQueue();
                sut.disconnect(cancelledError);
                expect(sut.lastState.value, const TypeMatcher<APIStateDisconnected>());
                expect(createdSockets.length, 1);
                sut.reconnect();
                await pumpEventQueue();
                expect(sut.lastState.value, const APIStateConnected());
                expect(createdSockets.length, 2);
                await sut.dispose();
            });

            group("stateMachine tests", () {
                test("initializes all states and starts out with the initialization state", () {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        const APIStateInitialized(),
                    ]));
                    expect(sut.lastState.value, const APIStateInitialized());
                });

                group("connected state", () {
                    group("policies", () {
                        test("$DontAutoConnectPolicy does nothing", () {
                            final socket = MockWebSocket();
                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    reducer,
                                ],
                                whenIsConnectedPolicy: const DontAutoConnectPolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                const APIStateInitialized(),
                                const APIStateLoading(),
                                const APIStateInitialized(),
                            ]));

                            final a = sut.lastData.subscribe();

                            a.cancel();
                        });

                        test("basic run through with $ConnectImmediatelyPolicy", () async {
                            final socket = MockWebSocket();
                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    reducer,
                                ],
                                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                ...apiInitLoadConnect,
                                const APIStateInitialized(),
                            ]));

                            final List<Event<Option<Either<AllErrors, SocketInterface>>>> events = [];

                            final d = sut.lastData.subscribeEvent((a) {
                                events.add(a);
                            });
                            await pumpEventQueue();

                            d.cancel();

                            expect(events, [
                                const NextEvent<Option<Either<AllErrors, SocketInterface>>>(None()),
                                NextEvent<Option<Either<AllErrors, SocketInterface>>>(Some(Right(socket))),
                                const CloseEvent<Option<Either<AllErrors, SocketInterface>>>(),
                            ]);
                        });

                        test("connects on first message when policy is $IsConnectedOnFirstMessagePolicy", () async {
                            final socket = MockWebSocket();
                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    reducer,
                                ],
                                whenIsConnectedPolicy: const IsConnectedOnFirstMessagePolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                ...apiInitLoadConnect,
                                const APIStateInitialized(),
                            ]));

                            final a = sut.lastData.subscribe((data) {});

                            await pumpEventQueue();

                            socket.kstream.add(const Some(Right("msg")));

                            await pumpEventQueue();

                            a.cancel();
                        });
                    });

                    group("reducer", () {
                        test("$ConnectedChannelReducer", () async {
                            final socket = MockWebSocket();
                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    ConnectedChannelReducer((a) => a.str.startsWith("1")),
                                ],
                                whenIsConnectedPolicy: const DontAutoConnectPolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                ...apiInitLoadConnect,
                                const APIStateInitialized(),
                            ]));

                            final a = sut.lastData.subscribe();

                            socket.kstream.add(const Some(Right("1")));

                            await pumpEventQueue();

                            a.cancel();
                        });

//                    test("$TimeoutChannelReducer", () async {
//                        final socket = MockWebSocket();
//                        sut = SocketEndpointImpl(
//                            socketBuilder: () => socket,
//                            reducers: [
//                                TimeoutChannelReducer(period: const Duration(milliseconds: 50), disconnectError: const TimeoutAPIError()),
//                            ],
//                            whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
//                        );
//
//                        expect(sut.lastState, emitsInOrder(<dynamic>[
//                            ...apiInitLoadConnect,
//                            const APIStateDisconnected(Some(TimeoutAPIError())),
//                        ]));
//
//                        expect(sut.lastData, emitsInOrder(<dynamic>[
//                            const TypeMatcher<Right<AllErrors, SocketInterface>>(),
//                            const Left(TimeoutAPIError()),
//                        ]));
//
//                        await Future<void>.delayed(const Duration(milliseconds: 30));
//                        socket.stream.add("1");
//
//                        await Future<void>.delayed(const Duration(milliseconds: 30));
//                        socket.stream.add("1");
//
//                        await Future<void>.delayed(const Duration(milliseconds: 30));
//                        socket.stream.add("1");
//
//                        await Future<void>.delayed(const Duration(milliseconds: 30));
//                        socket.stream.add("1");
//
//                        expect(sut.lastState, emitsInOrder(<dynamic>[
//                            const APIStateConnected(),
//                            const APIStateDisconnected(Some(TimeoutAPIError())),
//                        ]));
//                    });

                        test("$MarcopoloChannelReducer", () async {
                            final socket = MockWebSocket();
                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    MarcopoloChannelReducer<dynamic>((str) =>
                                    str.str.startsWith("a")
                                    ? Some(str.str.length.toString())
                                    : const None(),
                                    ),
                                ],
                                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                ...apiInitLoadConnect,
                                const APIStateInitialized(),
                            ]));

                            final sub = sut.lastData.subscribe();

                            expect(socket.msgs.wave.asStream(), emitsInOrder(<dynamic>[
                                null,
                                "1",
                                "3",
                                "4",
                            ]));

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("1")));

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("a")));

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("aaa")));

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("aaaa")));

                            await pumpEventQueue();

                            sub.cancel();
                        });

                        test("$ModelCallbackChannelReducer", () async {
                            final socket = MockWebSocket();
                            // ignore: close_sinks
                            final Signal<Option<int>> s = Signal(const None());

                            sut = SocketEndpointImpl(
                                socketCreator: () async => socket,
                                reducers: [
                                    ModelCallbackChannelReducer<int>(
                                        converter: (str) => str.str.length,
                                        callback: (a) => s.add(Some(a)),
                                    ),
                                ],
                                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                            );

                            expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                                ...apiInitLoadConnect,
                                const APIStateInitialized(),
                            ]));

                            expect(s.wave.asStream(), emitsInOrder(<dynamic>[
                                const None<int>(),
                                const Some(3),
                                const Some(4),
                            ]));

                            final sub = sut.lastData.subscribe();

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("aaa")));

                            await Future<void>.delayed(const Duration(milliseconds: 30));
                            socket.kstream.add(const Some(Right("aaaa")));

                            await pumpEventQueue();

                            sub.cancel();
                        });
                    });
                });
            });

            group("socketData", () {
                test("simple test", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                    ]));

                    sut.lastData.subscribe((a) {});

                    expect(sut.socketData.asStream(), emitsInOrder(<dynamic>[
                        const None<StringConversion>(),
                        Some(StringConversion("123")),
                        emitsDone,
                    ]));

                    await Future<void>.delayed(const Duration(milliseconds: 150));
                    createdSockets.first.kstream.add(const Some(Right("123")));
                    await Future<void>.delayed(const Duration(milliseconds: 150));
                    await sut.dispose();
                });

                test("test if socketData works through multiple reconnects of the underlying socket", () async {
                    MockWebSocket _socket;

                    int socketsCreated = 0;

                    final _sut = SocketEndpointImpl(
                        socketCreator: () async {
                            socketsCreated++;
                            return _socket = MockWebSocket();
                        },
                        reducers: [],
                        whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                    );

                    expect(_sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        APIStateDisconnected(some(unknownHintedAPIError("error"))),
                        ...apiInitLoadConnect,
                    ]));

                    _sut.lastData.subscribe();

                    expect(_sut.socketData.asStream(), emitsInOrder(<dynamic>[
                        const None<StringConversion>(),
                        Some(StringConversion("123")),
                        Some(StringConversion("123")),
                        emitsDone,
                    ]));

                    await Future<void>.delayed(const Duration(milliseconds: 150));

                    _socket.kstream.add(const Some(Right("123")));

                    _socket.kstream.add(Some(Left<AllErrors, String>(unknownHintedAPIError("error"))));

                    expect(socketsCreated, 1);

                    await Future<void>.delayed(const Duration(milliseconds: 20));

                    _sut.reconnect();

                    expect(socketsCreated, 2);

                    _socket.kstream.add(const Some(Right("123")));

                    await Future<void>.delayed(const Duration(milliseconds: 150));

                    expect(socketsCreated, 2);

                    await _sut.dispose();
                });
            });

            group("lastState & lastData", () {
                test("smoke test", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        const APIStateDisconnected(Some(cancelledError)),
                        emitsDone
                    ]));

                    expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<Either<AllErrors, SocketInterface>>(),
                        const TypeMatcher<Some<Either<AllErrors, SocketInterface>>>(),
                        emitsDone,
                    ]));

                    sut.lastData.subscribe();

                    await Future<void>.delayed(const Duration(milliseconds: 150));

                    await sut.dispose();
                });

                test("error thrown", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                        APIStateDisconnected(some(statusError(123))),
                    ]));

                    expect(sut.lastData.asStream(), emitsInOrder(<dynamic>[
                        const None<Either<AllErrors, SocketInterface>>(),
                        const TypeMatcher<Some<Either<AllErrors, SocketInterface>>>(),
                        Some(Left<dynamic, dynamic>(statusError(123))),
                        emitsDone
                    ]));

                    await Future<void>.delayed(const Duration(milliseconds: 150));

                    createdSockets.first.kstream.add(Some(Left(statusError(123))));

                    await Future<void>.delayed(const Duration(milliseconds: 50));

                    await sut.dispose();
                });
            });

            group("sendMessage", () {
                test("sendMessage working", () async {
                    expect(sut.lastState.asStream(), emitsInOrder(<dynamic>[
                        ...apiInitLoadConnect,
                    ]));

                    sut.lastData.subscribe();

                    expect(createdSockets.first.msgs.wave.skip(1).asStream(), emitsInOrder(<dynamic>[
                        "a",
                    ]));

                    await pumpEventQueue();

                    sut.sendMessage("a");
                });
            });
        });
    });
}

class TestModel extends SmartJSON<TestModel> {

    final SmartString a = SmartString("aa");

    final SmartString b = SmartString("bb");

    @override
    TestModel creator() => TestModel();

    @override
    JSONPrototypeCodec<TestModel> tactic() =>
        JSONMapTactic.ignoreMissing({"a": a, "b": b});
}