import 'dart:async';
import 'dart:convert';

import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:collection/collection.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

import '../../../../../wave_test.dart';

/// Integration tests for [SocketEndpointImpl] & [SocketChannelServiceEndpoint]
/// as well as various [SocketChannelReducer]s
///
/// All wrapped in an APIService as is should normally be
void main() {
  test("$_WebSocketTestRoutine1 test", () async {
    final websocket = _WebSocketTestRoutine1(
        "wss://test.test.com:443", const Duration(milliseconds: 100));

    websocket.startRoutine();

    websocket.stream.subscribe((_) {});

    /// websocket.stream.listen(print);

    expect(
      websocket.stream.asStream(),
      emitsInOrder(<dynamic>[
        const None<Either>(),
        const Some(Right<Object, String>('["BTCETH",true]')),
        const Some(
            Right<Object, String>('["BTCETH",{"price":"1.0","volume":"2.0"}]')),
        const Some(Right<Object, String>('["BTCDOGE",true]')),
        const Some(
            Right<Object, String>('["BTCETH",{"price":"3.0","volume":"4.0"}]')),
        const Some(Right<Object, String>(
            '["BTCDOGE",{"price":"1000.0","volume":"2000.0"}]')),
        const Some(Right<Object, String>('["BTCETH",false]')),
        const Some(Right<Object, String>(
            '["BTCDOGE",{"price":"3000.0","volume":"4000.0"}]')),
        const Some(Right<Object, String>('["BTCDOGE",false]')),
      ]),
    );

    websocket.add(subunsubMessage("BTCETH", true));
    await Future<void>.delayed(const Duration(milliseconds: 125));

    websocket.add(subunsubMessage("BTCDOGE", true));
    await Future<void>.delayed(const Duration(milliseconds: 125));

    websocket.add(subunsubMessage("BTCETH", false));
    await Future<void>.delayed(const Duration(milliseconds: 125));

    websocket.add(subunsubMessage("BTCDOGE", false));
  });

  test("socket only test", () async {
    int socketsCreated = 0;

    final SocketEndpointImpl socket = SocketEndpointImpl(
      socketCreator: () async {
        socketsCreated++;
        return _WebSocketTestRoutine1("wss://test.test.com:443")
          ..startRoutine();
      },
      whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
    );

    expect(socket.lastState.asStream(), emitsInOrder(<dynamic>[
      ...apiInitLoadConnect,
      const APIStateInitialized(),
    ]));

    final sub = socket.lastData.subscribe();

    await pumpEventQueue();
    sub.cancel();

    expect(socketsCreated, 1);
  });

  group("ideal conditions with no errors", () {
    test("simple test", () async {
      final _TestAPIService service = _TestAPIService();

      expect(service
          .defaultForeignSocket()
          .lastState
          .asStream(), emitsInOrder(<dynamic>[
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(CancelledAPIError())),
      ]));

      expect(service
          .tickerUpdate("BTCETH")
          .lastState
          .asStream(), emitsInOrder(<dynamic>[
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(CancelledAPIError())),
      ]));

      expect(service
          .tickerUpdate("BTCDOGE")
          .lastState
          .asStream(), emitsInOrder(<dynamic>[
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(CancelledAPIError())),
      ]));

      expect(service
          .tickerUpdate("BTCETH")
          .lastData
          .asStream(), emitsInOrder(<dynamic>[
        const None<Either>(),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        Some(Right<dynamic, dynamic>(
            OptionalParsedResult<TestModel>.raw(Some(Right(TestModel()
              ..price.safeSet("1.0")
              ..volume.safeSet("2.0")))))),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        Some(Right<dynamic, dynamic>(
            OptionalParsedResult<TestModel>.raw(Some(Right(TestModel()
              ..price.safeSet("3.0")
              ..volume.safeSet("4.0")))))),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
      ]));

      expect(service
          .tickerUpdate("BTCDOGE")
          .lastData
          .asStream(), emitsInOrder(<dynamic>[
        const None<Either>(),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        Some(Right<dynamic, dynamic>(
            OptionalParsedResult<TestModel>.raw(Some(Right(TestModel()
              ..price.safeSet("1000.0")
              ..volume.safeSet("2000.0")))))),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        Some(Right<dynamic, dynamic>(
            OptionalParsedResult<TestModel>.raw(Some(Right(TestModel()
              ..price.safeSet("3000.0")
              ..volume.safeSet("4000.0")))))),
      ]));

      await Future<void>.delayed(const Duration(milliseconds: 500));

      await service.dispose();

      expect(service.socketsCreated, 1);
    });

    test(
        "tests whether the disposal of the service correctly unsubs the endpoints", () async {
      final socket = _WebSocketTestRoutine1("wss://test.test.com:443");

      final _TestAPIService service = _TestAPIService(() => socket..startRoutine());

      expect(service
          .tickerUpdate("BTCETH")
          .lastState
          .asStream(), emitsInOrder(<dynamic>[
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(CancelledAPIError())),
        emitsDone,
      ]));

      expect(socket.receivedMessages.wave.asStream(), emitsInOrder(<dynamic>[
        null,
        subunsubMessage("BTCETH", true),
        subunsubMessage("BTCETH", false),
      ]));

      expect(service
          .tickerUpdate("BTCETH")
          .lastData
          .asStream(), emitsInOrder(<dynamic>[
        const None<Either>(),
        const Some(
            Right<dynamic, dynamic>(OptionalParsedResult<TestModel>.none())),
        Some(Right<dynamic, dynamic>(
            OptionalParsedResult<TestModel>.raw(Some(Right(TestModel()
              ..price.safeSet("1.0")
              ..volume.safeSet("2.0")))))),
        emitsDone,
      ]));

      await Future<void>.delayed(const Duration(milliseconds: 60));

      await service.dispose();
    });
  });

  group("with errors", () {
    test("no internet error thrown once and only test socket", () async {
      _WebSocketTestRoutine1 socket;

      final _TestAPIService service = _TestAPIService(() {
        socket = _WebSocketTestRoutine1("wss://test.test.com:443");
        return socket..startRoutine();
      });

      expect(service
          .defaultForeignSocket()
          .lastState
          .asStream(), emitsInOrder(<dynamic>[
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(NoInternetAPIError())),
        ...apiInitLoadConnect,
        const APIStateDisconnected(Some(CancelledAPIError())),
      ]));

      expect(service
          .defaultForeignSocket()
          .lastData
          .asStream(), emitsInOrder(<dynamic>[
        const None<Either>(),
            (dynamic a) =>
        a == Some<Either<AllErrors, _WebSocketTestRoutine1>>(Right(socket)),
        const Some(Left<dynamic, dynamic>(NoInternetAPIError())),
            (dynamic a) =>
        a == Some<Either<AllErrors, _WebSocketTestRoutine1>>(Right(socket)),
      ]));
      await Future<void>.delayed(const Duration(milliseconds: 200));

      socket._stream.add(const Some(Left(NoInternetAPIError())));
      await Future<void>.delayed(const Duration(milliseconds: 200));
      service.kdefaultForeignSocket.reconnect();
      await Future<void>.delayed(const Duration(milliseconds: 200));

      await service.dispose();

      expect(service.socketsCreated, 2);
    });

//        test("no internet error thrown once, test one channel", () async {
//            _RawTestWebSocket activeSocket;
//
//            List<_RawTestWebSocket> queue = [
//                _WebSocketTestRoutine2("wss://test.test.com:443"),
//                _WebSocketTestRoutine1("wss://test.test.com:443"),
//            ];
//
//            _TestAPIService service = _TestAPIService(() {
//                activeSocket = queue.removeAt(0);
//                return activeSocket..startRoutine();
//            });
//
//            expect(
//                service
//                    .tickerUpdate("BTCETH")
//                    .lastState
//                    .asStream(),
//                emitsInOrder(<dynamic>[
//                    ...apiInitLoadConnect,
//                    const APIStateDisconnected(Some(NoInternetAPIError())),
//                    apiInitialized,
//                    apiLoading,
//                    const APIStateDisconnected(Some(CancelledAPIError())),
//                    emitsDone,
//                ]),
//            );
//
//            expect(
//                service
//                    .defaultForeignSocket()
//                    .lastState
//                    .asStream(),
//                emitsInOrder(<dynamic>[
//                    ...apiInitLoadConnect,
//                    apiInitialized, /// should be below?
//                    const APIStateDisconnected(Some(NoInternetAPIError())),
//                    apiLoading,
//                    const APIStateDisconnected(Some(CancelledAPIError())),
//                    emitsDone,
//                ]),
//            );
//
//            service
//                .tickerUpdate("BTCETH")
//                .lastData
//                .subscribe();
//
//            await Future<void>.delayed(const Duration(milliseconds: 500));
//
//            expect(service.socketsCreated, 1);
//
//            service.tickerUpdate("BTCETH").reconnect();
//
//            expect(service.socketsCreated, 2);
//
//            await pumpEventQueue();
//
//            await service.dispose();
//        });
  });

  test("heartbeat test & stream closing unexpectedly", () async {
    final _WebSocketTestRoutinePing webSocketRoutine = _WebSocketTestRoutinePing(
      const Duration(milliseconds: 50),
      const Duration(milliseconds: 15),
    );

    final _TestAPIService service = _TestAPIService(() => webSocketRoutine);

    final WaveTester testReceivedMessages = WaveTester<String>(
        webSocketRoutine.receivedMessages.wave);

    expect(service
        .defaultForeignSocket()
        .lastState
        .asStream(), emitsInOrder(<dynamic>[
      const APIStateInitialized(),
      const APIStateLoading(),
      const APIStateConnected(),
      const APIStateDisconnected(Some(cancelledError))
    ]));

    expect(webSocketRoutine.stream.asStream(), emitsInOrder(<dynamic>[
      const None<Either>(),
      const Some(Right<AllErrors, String>("ping")),
      const Some(Right<AllErrors, String>("ping")),
      const Some(Right<AllErrors, String>("ping")),
      const Some(Right<AllErrors, String>("ping")),
      const Some(Right<AllErrors, String>("ping")),
    ]));
    service
        .tickerUpdate("BTCETH")
        .lastData
        .subscribe();

    webSocketRoutine.startRoutine();

    await Future<void>.delayed(const Duration(milliseconds: 400));

    testReceivedMessages.expectCancelDidEmitTAndEnd(<String>[
      null,
      subunsubMessage("BTCETH", true),
      "pong",
      "pong",
      "pong",
      "pong",
      "pong",
    ]);

    await service.dispose();
  });
}

/// For testing purposes, data messages meant for a channel are expected to have the following form:
/// [channel, content (Map)]
///
/// subscription to a channel is confirmed by
/// [channel, true]
///
/// unsubscription to a channel is confirmed by
/// [channel, false]
class _TestAPISocketDefinition extends APIWebSocketDescription {

  const _TestAPISocketDefinition() : super(
    host: "test.test.com",
  );

  SingleChannelSocketDefinition<OptionalParsedResult<TestModel>> ticker(
      String pair) =>
      single(
        pair,
        preferences: (name, id) =>
            SocketChannelPreferences(
              sub: subunsubMessage(pair, true),
              unsub: subunsubMessage(pair, false),
              reducers: [
                ConnectedChannelReducer(
                      (str) =>
                      OptionalParsedResult<bool>(
                          source: str,
                          converter: (a) =>
                              const ListEquality<dynamic>().equals(
                                  str.cachedJson() as List,
                                  <dynamic>[name, true]))
                          .value
                          .map((a) => a.fold((a) => false, (a) => a))
                          .getOrElse(() => false),
                ),
              ],
              whenIsConnectedPolicy: const DontAutoConnectPolicy(),
            ),
        converter: (str, info) =>
            OptionalParsedResult(
              source: str,
              preInteresting: [
                PreInterestCheck(
                        (s) =>
                    s.cachedJson() is List
                        && s.cachedJson()[0] == info.name
                        && s.cachedJson()[1] is Map,
                    "filter for channelName"
                ),
              ],
              validations: [
                ValidationCheck(
                      (t) => t.price.nullVal() != null,
                  "price needs to be present",
                ),
                ValidationCheck(
                      (t) => t.volume.nullVal() != null,
                  "volume needs to be present",
                ),
              ],
              converter: (str) => TestModel().from(str.cachedJson()[1]),
            ),
      );
}

class TestModel extends SmartJSON<TestModel> {

  final SmartLazyDecimal price = SmartLazyDecimal(null);

  final SmartLazyDecimal volume = SmartLazyDecimal(null);

  @override
  TestModel creator() => TestModel();

  @override
  JSONPrototypeCodec<TestModel> tactic() =>
      JSONMapTactic.ignoreMissing({
        "price": price,
        "volume": volume,
      });
}

class _TestAPIService extends APIService {

  _TestAPISocketDefinition socketDef = const _TestAPISocketDefinition();

  _RawTestWebSocket Function() altSocket;

  _TestAPIService([this.altSocket]) {
    kdefaultForeignSocket = SocketEndpointImpl(
      socketCreator: () async {
        socketsCreated++;
        return altSocket?.call() ?? (_WebSocketTestRoutine1(socketDef.url)
          ..startRoutine());
      },
      whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
      reducers: [
        MarcopoloChannelReducer<dynamic>((str) =>
        (str.str == "ping") ?
        some("pong") : none()),
      ],
    );
  }

  int socketsCreated = 0;

  SocketEndpointInterface kdefaultForeignSocket;

  @override
  SocketEndpointInterface defaultForeignSocket() => kdefaultForeignSocket;

  ServiceEndpoint<OptionalParsedResult<TestModel>> tickerUpdate(String pair) =>
      socketForeign(
        cacheId: "tickerUpdate$pair",
        channel: socketDef.ticker(pair),
      );
}

String subunsubMessage(String channel, bool subscribe) =>
    json.encode(["subscribe", channel, subscribe]);

class _RawTestWebSocket extends SocketInterface {

  @override
  Future close() async {
    _stream.close();
  }

  @override
  void add(String str) {
    receivedMessages.add(str);
  }

  final Signal<Option<Either<AllErrors, String>>> _stream = Signal(
      const None<Either<AllErrors, String>>());

  @override
  Wave<Option<Either<AllErrors, String>>> get stream => _stream.wave;

  /// MARK: for testing

  final String url;

  final Signal<String> receivedMessages = Signal(null);

  _RawTestWebSocket(this.url);

  void startRoutine() =>
      throw Exception("use subclasses to test the websocket");
}

const String btcdogeChannel = "BTCDOGE";

const String btcethChannel = "BTCETH";

/// A simple routine
class _WebSocketTestRoutine1 extends _RawTestWebSocket {

  final Duration waitDelay;

  _WebSocketTestRoutine1(String url,
      [this.waitDelay = const Duration(milliseconds: 30)])
      : assert(url == "wss://test.test.com:443"),
        super(url);

  Future waitALittle() => Future<void>.delayed(waitDelay);

  String _sendValidChannelMessage(String channel, dynamic content) =>
      json.encode(<dynamic>[channel, content]);

  String _sendConnectedChannelMessage(String channel) =>
      json.encode([channel, true]);

  String _sendDisconnectedChannelMessage(String channel) =>
      json.encode([channel, false]);

  StreamSubscription<String> btcethSubscription;

  Stream<String> btcEth() async* {
    yield _sendConnectedChannelMessage(btcethChannel);

    await waitALittle();

    while (btcethSubscription != null) {
      yield _sendValidChannelMessage(btcethChannel, (TestModel()
        ..price.safeSet("1.0")
        ..volume.safeSet("2.0")).to());
      await waitALittle();
      yield _sendValidChannelMessage(btcethChannel, (TestModel()
        ..price.safeSet("3.0")
        ..volume.safeSet("4.0")).to());
      await waitALittle();
    }
  }

  StreamSubscription<String> btcdogeSubscription;

  Stream<String> btcDoge() async* {
    yield _sendConnectedChannelMessage(btcdogeChannel);

    await waitALittle();

    while (btcdogeSubscription != null) {
      yield _sendValidChannelMessage(btcdogeChannel, (TestModel()
        ..price.safeSet("1000.0")
        ..volume.safeSet("2000.0")).to());
      await waitALittle();
      yield _sendValidChannelMessage(btcdogeChannel, (TestModel()
        ..price.safeSet("3000.0")
        ..volume.safeSet("4000.0")).to());
      await waitALittle();
    }
  }

  @override
  void startRoutine() {
    receivedMessages.wave.subscribe((a) {
      if (a == subunsubMessage(btcethChannel, true)) {
        btcethSubscription = btcEth().listen((a) {
          if (!_stream.isClosed) {
            _stream.add(Some(Right(a)));
          }
        });
      }

      if (a == subunsubMessage(btcethChannel, false)) {
        if (btcethSubscription != null) {
          if (!_stream.isClosed) {
            _stream.add(
                Some(Right(_sendDisconnectedChannelMessage(btcethChannel))));
          }
          btcethSubscription.cancel();
        }
      }

      if (a == subunsubMessage(btcdogeChannel, true)) {
        btcdogeSubscription = btcDoge().listen((a) {
          if (!_stream.isClosed) {
            _stream.add(Some(Right(a)));
          }
        });
      }

      if (a == subunsubMessage(btcdogeChannel, false)) {
        if (btcethSubscription != null) {
          if (!_stream.isClosed) {
            _stream.add(
                Some(Right(_sendDisconnectedChannelMessage(btcdogeChannel))));
          }
          btcdogeSubscription.cancel();
        }
      }
    });
  }
}

///// A simple routine that gives an error
//class _WebSocketTestRoutine2 extends _RawTestWebSocket {
//
//    final Duration waitDelay;
//
//    _WebSocketTestRoutine2(String url, [this.waitDelay = const Duration(milliseconds: 30)])
//        : assert(url == "wss://test.test.com:443"),
//            super(url);
//
//    Future waitALittle() => Future<void>.delayed(waitDelay);
//
//    String _sendValidChannelMessage(String channel, dynamic content) => json.encode(<dynamic>[channel, content]);
//
//    String _sendConnectedChannelMessage(String channel) => json.encode([channel, true]);
//
//    String _sendDisconnectedChannelMessage(String channel) => json.encode([channel, false]);
//
//    Disposable btcethSubscription;
//
//    Wave<String> btcEth() {
//        return Wave.just(_sendConnectedChannelMessage(btcethChannel));
//    }
//
//    Disposable btcdogeSubscription;
//
//    Wave<String> btcDoge() {
//        return Wave.just(_sendConnectedChannelMessage(btcdogeChannel));
//    }
//
//    @override
//    void startRoutine() async {
//        receivedMessages.wave.subscribe((a) {
//            if (a == subunsubMessage(btcethChannel, true)) {
//                btcethSubscription = btcEth().subscribe((a) => _stream.add(Some(Right(a))));
//            }
//
//            if (a == subunsubMessage(btcethChannel, false)) {
//                if (btcethSubscription != null) {
//                    if (!_stream.isClosed)
//                        _stream.add(Some(Right(_sendDisconnectedChannelMessage(btcethChannel))));
//                    btcethSubscription.cancel();
//                }
//            }
//
//            if (a == subunsubMessage(btcdogeChannel, true)) {
//                btcdogeSubscription = btcDoge().subscribe((a) => _stream.add(Some(Right(a))));
//            }
//
//            if (a == subunsubMessage(btcdogeChannel, false)) {
//                if (btcethSubscription != null) {
//                    if (!_stream.isClosed)
//                        _stream.add(Some(Right(_sendDisconnectedChannelMessage(btcdogeChannel))));
//                    btcdogeSubscription.cancel();
//                }
//            }
//        });
//
//        await Future<void>.delayed(const Duration(milliseconds: 200));
//
//        if (!_stream.isClosed)
//            _stream.add(const Some(Left(NoInternetAPIError())));
//    }
//}

/// A simple routine that gives an error
class _WebSocketTestRoutinePing extends _RawTestWebSocket {

  final Duration pingPongDelay;
  final Duration waitDelay;

  _WebSocketTestRoutinePing(this.pingPongDelay, this.waitDelay)
      : super("not needed");

  Timer t;

  int times = 5;

  void setTimer() {
    times--;
    if (!_stream.isClosed) {
      t = Timer(pingPongDelay, _stream.close);
      _stream.add(const Some(Right("ping")));
    }
  }

  @override
  void startRoutine() {
    setTimer();
    receivedMessages.wave.subscribe((a) async {
      if (a == "pong") {
        if (times > 0) {
          setTimer();
        }
      }
    });
  }
}