import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

void main() {
    group("$APIService", () {
        test("cachedAccess", () {
            final _TestAPIService sut = _TestAPIService();
            final _MockServiceEndpoint endpoint = _MockServiceEndpoint();
            sut.cachedAccess<dynamic>("a", () => endpoint);
            expect(sut.cache.values.length, 1);
            expect(sut.cache.keys.first, "a");
            sut.cachedAccess<dynamic>("a", () => endpoint);
            expect(sut.cache.values.length, 1);
            expect(sut.cache.keys.first, "a");
            sut.cachedAccess<dynamic>("b", () => endpoint);
            expect(sut.cache.values.length, 2);
            expect(sut.cache.keys, ["a", "b"]);
        });

        group("rest", () {
            test("is cached", () {
                final _TestAPIService sut = _TestAPIService();
                final _MockRAWModelUpdate modelUpdate = _MockRAWModelUpdate();
                sut.rest("a", modelUpdate);
                expect(sut.cache.values.length, 1);
                expect(sut.cache.keys.first, "a");
            });
        });

        group("defaultForeignSocket", () {
            test("by default null", () {
                final _TestAPIService sut = _TestAPIService();
                expect(sut.defaultForeignSocket(), isNull);
            });
        });

        group("socketForeign", () {
            final _MockSocketChannelPreferences prefs = _MockSocketChannelPreferences();

            final SingleChannelSocketDefinition channel = SingleChannelSocketDefinition<dynamic>(
                channelName: "",
                buildUrl: (_, __, ___) => "",
                converter: (_, __) => null,
                description: null,
                id: "",
                preferences: prefs,
            );

            when(prefs.reducers).thenReturn([]);

            test("is cached", () {
                final _TestAPIService sut = _TestAPIService();
                sut.socketForeign<dynamic>(
                    cacheId: "a",
                    socket: () => _MockSocketEndpoint(),
                    channel: channel,
                );
                expect(sut.cache.values.length, 1);
                expect(sut.cache.keys.first, "a");
            });

            test("uses default socket", () {
                final _TestAPIServiceWithForeignSocket service = _TestAPIServiceWithForeignSocket();

                /// access once to cache it
                service.socketForeign<dynamic>(cacheId: "", channel: channel);
                expect(
                    (service.socketForeign<dynamic>(cacheId: "", channel: channel) as SocketChannelServiceEndpoint).socket(null),
                    service.socketEndpoint);
            });

            test("uses passed socket", () {
                final _TestAPIServiceWithForeignSocket service = _TestAPIServiceWithForeignSocket();
                final socket = _MockSocketEndpoint();
                expect(identical(
                    (service.socketForeign<dynamic>(
                        cacheId: "", socket: () => socket, channel: channel) as SocketChannelServiceEndpoint).socket(null),
                    socket,
                ), true);
            });

            test("throws if neither default socket set nor socket passed", () {
                final _TestAPIService service = _TestAPIService();
                expect(() =>
                    (service.socketForeign<dynamic>(cacheId: "", channel: channel) as SocketChannelServiceEndpoint).socket(null),
                    throwsA(const TypeMatcher<Exception>()));
            });
        });

        group("socketOwn", () {
            final _MockSocketChannelPreferences prefs = _MockSocketChannelPreferences();

            when(prefs.whenIsConnectedPolicy).thenReturn(const ConnectImmediatelyPolicy());
            when(prefs.onReceivedSocket).thenReturn(Some((a) {}));
            when(prefs.reducers).thenReturn([]);

            final SingleChannelSocketDefinition def = SingleChannelSocketDefinition<dynamic>(
                channelName: "",
                buildUrl: (_, __, ___) => "",
                converter: (_, __) => null,
                description: null,
                id: "",
                preferences: prefs,
            );

            test("is cached", () {
                final _TestAPIService sut = _TestAPIService();

                sut.socketOwn<dynamic>(
                    cacheId: "a",
                    socketBuilder: (url) => _MockSocketEndpoint(),
                    channel: def,
                );
                expect(sut.cache.values.length, 1);
                expect(sut.cache.keys.first, "a");
            });

            test("correct url given", () {
                final _TestAPIService sut = _TestAPIService();
                final _MockSingleChannelSocketDefinition channel = _MockSingleChannelSocketDefinition<dynamic>();

                when(channel.preferences).thenReturn(prefs);
                when(channel.url).thenReturn("abc");
                when(channel.converter).thenReturn((a, b) => a.str);

                final s = sut.socketOwn<dynamic>(
                    cacheId: "a",
                    socketBuilder: (url) => _MockSocketEndpoint(url),
                    channel: channel,
                );

                expect((s as SocketChannelServiceEndpoint).channelUrl.url, "abc");
            });

            test("converter", () async {
                final _TestAPIService sut = _TestAPIService();

                _MockWebSocket wsocket;

                final SocketEndpointImpl socket = SocketEndpointImpl(
                    socketCreator: () async => wsocket = _MockWebSocket(),
                    whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                );

                final s = sut.socketOwn<int>(
                    cacheId: "a",
                    socketBuilder: (_) => socket,
                    channel: SingleChannelSocketDefinition<int>(
                        buildUrl: (_, __, ___) => "a",
                        channelName: "abc",
                        converter: (StringConversion a, SocketChannelURLAndMoreProvider b) => a.str.length,
                        description: null,
                        id: "id",
                        preferences: prefs,
                    ),
                );

                expect(s.lastData.asStream(), emitsInOrder(const <dynamic>[
                    None<Either>(),
                    Some(Right<dynamic, dynamic>(3)),
                    Some(Right<dynamic, dynamic>(5)),
                    Some(Right<dynamic, dynamic>(7)),
                ]));

                wsocket._stream.add(const Some(Right("123")));
                await pumpEventQueue();
                wsocket._stream.add(const Some(Right("aaaaa")));
                await pumpEventQueue();
                wsocket._stream.add(const Some(Right("abc1234")));
            });
        });

        test("dispose all service endpoints", () async {
            final _TestAPIService sut = _TestAPIService();
            final _MockServiceEndpoint endpoint = _MockServiceEndpoint();
            when(endpoint.dispose()).thenAnswer((_) async {});
            sut.cachedAccess<dynamic>("aaa", () => endpoint);
            await sut.dispose();
            verify(endpoint.dispose()).called(1);
        });

        test("dispose a default foreign socket", () async {
            final _TestAPIServiceWithForeignSocket sut = _TestAPIServiceWithForeignSocket();
            when(sut.socketEndpoint.dispose()).thenAnswer((_) async {});
            await sut.dispose();
            verify(sut.socketEndpoint.dispose()).called(1);
        });
    });
}

class _TestAPIService extends APIService {}

class _TestAPIServiceWithForeignSocket extends APIService {

    _MockSocketEndpoint socketEndpoint = _MockSocketEndpoint();

    @override
    SocketEndpointImpl defaultForeignSocket() => socketEndpoint;
}

class _MockSingleChannelSocketDefinition<T> extends Mock implements SingleChannelSocketDefinition<T> {}

class _MockServiceEndpoint extends Mock implements ServiceEndpoint<dynamic> {}

class _MockSocketEndpoint extends Mock implements SocketEndpointImpl {
    final String url;

    _MockSocketEndpoint([this.url]);
}

class _MockWebSocket extends SocketInterface {

    final Signal<String> msgs = Signal<String>("");

    @override
    Future close() async {
        msgs.close();
        _stream.close();
    }

    @override
    void add(String str) => msgs.add(str);

    final Signal<Option<Either<AllErrors, String>>> _stream = Signal(const None());

    @override
    Wave<Option<Either<AllErrors, String>>> get stream => _stream.wave;
}

class _MockSocketChannelPreferences extends Mock implements SocketChannelPreferences {}

class _MockRAWModelUpdate extends Mock implements RAWModelUpdate<TestModel, String> {}

class TestModel extends SmartJSON<TestModel> {
    final SmartString a = SmartString("aa");
    final SmartString b = SmartString("bb");

    @override
    TestModel creator() => TestModel();

    @override
    JSONPrototypeCodec<TestModel> tactic() =>
        JSONMapTactic.ignoreMissing({"a": a, "b": b});
}