import 'package:bird/signal_wave.dart';
import 'package:bird/smart_api.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

void main() {
    group("$SocketInterface", () {
        /// Nothing to test
    });

    group("$APIWebSocketDescription", () {
        final _TestSocketDescription desc = _TestSocketDescription(_MockWebSocket());

        test("host", () {
            expect(desc.host, "demos.kaazing.com");
        });

        test("path", () {
            expect(desc.path, "echo");
        });

        test("port", () {
            expect(desc.port, 443);
        });

        test("defaultSingleUrlBuilder", () {
            expect(desc.defaultSingleUrlBuilder("-", "/", desc), desc.buildUriFor());
        });

        test("defaultChannelPreferences", () {
            expect(desc
                .defaultChannelPreferences("-", "-")
                .onLostSocket, none<dynamic>());
            expect(desc
                .defaultChannelPreferences("-", "-")
                .onReceivedSocket, none<dynamic>());
            expect(desc
                .defaultChannelPreferences("-", "-")
                .reducers, <dynamic>[]);
        });

        group("single", () {
            test("channelName is passed", () {
                expect(
                    desc
                        .single(
                        "channelName",
                        converter: (conversion, data) {
                            return "";
                        },
                    )
                        .channelName, "channelName",
                );
            });

            test("channel name is the id on default", () {
                expect(
                    desc
                        .single(
                        "channelName",
                        converter: (conversion, data) {
                            return "";
                        },
                    )
                        .id, "channelName",
                );
            });

            test("idBuilder changes id", () {
                expect(
                    desc
                        .single(
                        "channelName",
                        idBuilder: (a) => "$a+",
                        converter: (conversion, data) {
                            return "";
                        },
                    )
                        .id, "channelName+",
                );
            });

            test("set preferences are passed", () {
                final SocketChannelPreferences a = SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy());
                expect(
                    identical(desc
                        .single(
                        "channelName",
                        preferences: (name, id) => a,
                        idBuilder: (a) => "$a+",
                        converter: (conversion, data) => "",
                    )
                        .preferences, a),
                    isTrue
                );
            });

            test("passed url builder is used", () {
                expect(
                    desc
                        .single(
                        "channelName",
                        urlBuilder: (a, b, c) => "ok",
                        converter: (conversion, data) => "",
                    )
                        .url,
                    "ok",
                );
            });

            test("default url builder is used", () {
                final _TestSocketDescriptionDefaultPrefs d = _TestSocketDescriptionDefaultPrefs(_MockWebSocket());

                expect(
                    d
                        .single(
                        "channelName",
                        converter: (conversion, data) => "",
                    )
                        .url,
                    "myURL",
                );
            });

            test("defaultPreferences are used", () {
                final SocketChannelPreferences p = SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy());
                final _TestSocketDescriptionDefaultPrefs d = _TestSocketDescriptionDefaultPrefs(_MockWebSocket());
                d.defPref = p;
                expect(
                    identical(d
                        .single(
                        "channelName",
                        idBuilder: (a) => "$a+",
                        converter: (conversion, data) => "",
                    )
                        .preferences, p),
                    isTrue
                );
            });
        });

        group("buildUriFor", () {
            test("alone", () {
                expect(
                    desc.buildUriFor(),
                    "wss://demos.kaazing.com:443/echo",
                );
            });

            test("with additiionalPath", () {
                expect(
                    desc.buildUriFor(additionalPath: "additional"),
                    "wss://demos.kaazing.com:443/echo/additional",
                );
            });

            test("ith string arguments", () {
                expect(
                    desc.buildUriFor(args: <String, dynamic>{"marco": "polo", "foo": "bar"}),
                    "wss://demos.kaazing.com:443/echo?marco=polo&foo=bar",
                );
            });

            test("with list arguments", () {
                expect(
                    desc.buildUriFor(args: <String, dynamic>{"marco": ["po", "lo"], "foo": ["bar", null, "bar"]}),
                    "wss://demos.kaazing.com:443/echo?marco=po&marco=lo&foo=bar&foo&foo=bar",
                );
            });

            test("with arguments and additionalPath", () {
                expect(
                    desc.buildUriFor(
                        additionalPath: "more",
                        args: <String, dynamic>{"marco": ["po", "lo"], "foo": ["bar", null, "bar"]},
                    ),
                    "wss://demos.kaazing.com:443/echo/more?marco=po&marco=lo&foo=bar&foo&foo=bar",
                );
            });
        });

        test("url", () {
            expect(desc.url, desc.buildUriFor());
        });
    });

    group("$SocketChannelPreferences", () {
        test("default parameters", () {
            expect(SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()).reducers, <dynamic>[]);
            expect(SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()).onReceivedSocket,
                none<dynamic>());
            expect(
                SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()).onLostSocket, none<dynamic>());
        });

        test("non default parameters", () {
            expect(SocketChannelPreferences(
                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                reducers: [
                    DebugSocketChannelReducer("-"),
                ],
            ).reducers.length, 1);
            expect(SocketChannelPreferences(
                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
                sub: "aa",
            ).onReceivedSocket.isSome(), true);

            String subMessage;
            String unsubMessage;

            SocketChannelPreferences(
                sub: "123",
                unsub: "321",
                whenIsConnectedPolicy: const ConnectImmediatelyPolicy(),
            )
                ..onReceivedSocket.fold(
                        () {},
                        (add) {
                        add((a) {
                            subMessage = a;
                        });
                    },
                )
                ..onLostSocket.fold(
                        () {},
                        (add) {
                        add((a) {
                            unsubMessage = a;
                        });
                    },
                )
            ;

            expect(subMessage, "123");
            expect(unsubMessage, "321");
        });
    });

    group("$SocketChannelUrlProvider", () {
        /// Nothing to test
    });

    group("$SocketChannelURLAndMoreProvider", () {
        /// Nothing to test
    });

    group("$SingleChannelSocketDefinition", () {
        test("name", () {
            final _TestSocketDescription a = _TestSocketDescription(_MockWebSocket());
            expect(SingleChannelSocketDefinition(
                description: a,
                channelName: "cname",
                buildUrl: (a, b, c) => "",
                converter: (a, b) => a.str,
                id: "id",
                preferences: SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()),
            ).name, "cname");
        });

        test("url", () {
            final _TestSocketDescription a = _TestSocketDescription(_MockWebSocket());
            expect(SingleChannelSocketDefinition(
                description: a,
                channelName: "cname",
                buildUrl: (a, b, c) => "$a$b",
                converter: (a, b) => a.str,
                id: "id",
                preferences: SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()),
            ).url, "cnameid");
        });

        test("thisConverter", () {
            final _TestSocketDescription a = _TestSocketDescription(_MockWebSocket());
            SingleChannelSocketDefinition d;

            d = SingleChannelSocketDefinition<dynamic>(
                description: a,
                channelName: "cname",
                buildUrl: (a, b, c) => "$a$b",
                converter: (a, b) {
                    expect(identical(d, b), true);
                    return a.str;
                },
                id: "id",
                preferences: SocketChannelPreferences(whenIsConnectedPolicy: const ConnectImmediatelyPolicy()),
            );

            expect(d.thisConverter(StringConversion("123")), "123");
        });
    });
}

class _TestSocketDescription extends APIWebSocketDescription {

    final _MockWebSocket socket;

    _TestSocketDescription(this.socket) : super(host: "demos.kaazing.com", path: "echo", port: 443);
}

class _TestSocketDescriptionDefaultPrefs extends APIWebSocketDescription {

    final _MockWebSocket socket;

    SocketChannelPreferences defPref;

    @override
    SocketChannelPreferences defaultChannelPreferences(String name, String defid) => defPref;

    @override
    String defaultSingleUrlBuilder(String channelName, String id, APIWebSocketDescription description) {
        return "myURL";
    }

    _TestSocketDescriptionDefaultPrefs(this.socket) : super(host: "demos.kaazing.com", path: "echo", port: 443);
}

class _MockWebSocket extends SocketInterface {

    List<String> msgs = [];

    @override
    void add(String str) => msgs.add(str);

    @override
    Future close() async => msgs.add("CLOSE");

    @override
    Wave<Option<Either<AllErrors, String>>> get stream {
        return Wave.timer(const Duration(milliseconds: 20), (a) => const Some(Right<AllErrors, String>("ping")));
    }
}