import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:decimal/decimal.dart';
import 'package:test/test.dart';

void main() {
  group("serialization_tactic.dart", () {
    final map = {"a": "b"};
    final list = [true, 1, 1.2, "hi"];
    final list2 = [false, 11, 2.2, "hello"];

    test("not null", () {
      expect(() =>
      JSONMapTactic<_MapNull>.raw(
          null, (a) => null, MapMissingBehaviour.ignore),
          throwsA(const TypeMatcher<AssertionError>()));
      expect(() =>
      JSONMapTactic<_MapNull>.raw({}, null, MapMissingBehaviour.ignore),
          throwsA(const TypeMatcher<AssertionError>()));
      expect(() => JSONMapTactic<_MapNull>.raw({}, (a) => null, null),
          throwsA(const TypeMatcher<AssertionError>()));
    });

    test("JSONMapTactic missing null", () {
      final m = _MapNull();
      expect(m.from(map).to(), {"a": "b", "x": null});
    });

    test("JSONMapTactic ignore null", () {
      final _MapIgnore m = _MapIgnore();
      expect(m.from(map).to(), {"a": "b", "x": "123"});
    });

    test("JSONMapTactic can deal with nulls", () {
      final _MapIgnore m = _MapIgnore();
      expect(m.from(null).to(), {"a": "123", "x": "123"});
      expect(m.from(null).toString(), '{"a":"123","x":"123"}');
    });

    final mapaslistmap = {
      "123.123": "234.234",
      "345.345": "456.456",
    };

    final _MapAsListNull m = _MapAsListNull().from(mapaslistmap);

    group("JSONMapAsListTactic", () {
      test("JSONMapAsListTactic fromJson keys", () {
        expect(_MapAsListNull()
            .from(mapaslistmap)
            .keys
            .to(), ["123.123", "345.345"]);
      });

      test("JSONMapAsListTactic fromJson values", () {
        expect(_MapAsListNull()
            .from(mapaslistmap)
            .values
            .to(), ["234.234", "456.456"]);
      });

      test("JSONMapAsListTactic keys toStrinng", () {
        expect(
            m.keys.toString(), json.encode(mapaslistmap.keys.toList()));
      });

      test("JSONMapAsListTactic values toString", () {
        expect(m.values.toString(),
            json.encode(mapaslistmap.values.toList()));
      });

      test("JSONMapAsListTactic fromJson toJson", () {
        expect(_MapAsListNull().from(mapaslistmap).to(), mapaslistmap);
      });

      test("JSONMapAsListTactic invalid value and key", () {
        expect(
            _MapAsListNull().from({null: "123.0", "123.1": null}).to(),
            {null: "123.0", "123.1": null});
      });
      test("asMap", () {
        expect(_MapAsListNull().from(mapaslistmap).tactic().asMap(), {
          Decimal.tryParse("123.123"): Decimal.tryParse("234.234"),
          Decimal.tryParse("345.345"): Decimal.tryParse("456.456"),
        });
      });
    });

    test("JSONListTactic", () {
      final _List2 m = _List2();
      final a = m.from(list);
      expect(a.to(), list);
      expect(a.a.nullVal(), true);
      expect(a.b.nullVal(), 1);
      expect(a.c.nullVal(), 1.2);
      expect(a.d.nullVal(), "hi");
    });

    test("JSONSingleTactic lists", () {
      final dynamic structure = [list, list2];
      final _SingleList m = _SingleList();
      expect(m.from(structure).to(), structure);
      expect(m.from([[true]]).to(), [[true, null, null, null]]);
    });

    test("Alternative map for fromJson", () {
      expect(_MapIgnoreAlternativeValues().from({"b": "HELLO"}).to(),
          {"a": "123"});
      expect(_MapIgnoreAlternativeValues().from({"a": "HELLO"}).to(),
          {"a": "HELLO"});
      expect(_MapIgnoreAlternativeValues().from(
          {"a": "HELLO", "A": "ITS ME"}).to(), {"a": "ITS ME"});
      expect(_MapIgnoreAlternativeValues().from({"A": "ITS ME"}).to(),
          {"a": "ITS ME"});
    });
  });
}

class _SingleList extends SmartJSON<_SingleList> {
  final SmartModelList<_List2> a = SmartModelList([], () => _List2());

  @override
  _SingleList creator() => _SingleList();

  @override
  JSONPrototypeCodec<_SingleList> tactic() => JSONSingleTactic(a);
}

class _List2 extends SmartJSON<_List2> {
  final SmartBool a = SmartBool(null);
  final SmartInt b = SmartInt(null);
  final SmartDouble c = SmartDouble(null);
  final SmartString d = SmartString(null);

  @override
  _List2 creator() => _List2();

  @override
  JSONPrototypeCodec<_List2> tactic() => JSONListTactic([a, b, c, d]);
}

class _MapNull extends SmartJSON<_MapNull> {
  final SmartString a = SmartString("123");
  final SmartString x = SmartString("123");

  @override
  _MapNull creator() => _MapNull();

  @override
  JSONPrototypeCodec<_MapNull> tactic() =>
      JSONMapTactic.nullOnMissing({"a": a, "x": x});
}

class _MapAsListNull extends SmartJSON<_MapAsListNull> {
  final SmartDecimalList keys = SmartDecimalList([]);
  final SmartDecimalList values = SmartDecimalList([]);

  @override
  _MapAsListNull creator() => _MapAsListNull();

  @override
  JSONMapAsListTactic<_MapAsListNull, Decimal, Decimal> tactic() =>
      JSONMapAsListTactic(keys, values);
}

class _MapIgnore extends SmartJSON<_MapIgnore> {
  final SmartString a = SmartString("123");
  final SmartString x = SmartString("123");

  @override
  _MapIgnore creator() => _MapIgnore();

  @override
  JSONPrototypeCodec<_MapIgnore> tactic() =>
      JSONMapTactic.ignoreMissing({"a": a, "x": x});
}

class _MapIgnoreAlternativeValues
    extends SmartJSON<_MapIgnoreAlternativeValues> {
  final SmartString a = SmartString("123");

  @override
  _MapIgnoreAlternativeValues creator() => _MapIgnoreAlternativeValues();

  @override
  JSONPrototypeCodec<_MapIgnoreAlternativeValues> tactic() =>
      JSONCanonicalMapTactic<
          _MapIgnoreAlternativeValues>.ignoreOnMissingCanonical(
          {"a": a, "A": a}, {"a": a});
}
